#  wfield - tools to analyse widefield data - svd utils 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *

class svd_pix_correlation():
    def __init__(self,U,SVT,dims = None, norm_svt=False):
        '''
        Local correlation using the decomposed matrices.
        '''
        normed = SVT.copy()
        if issparse(U):
            if dims is None:
                raise ValueError('Supply dims when using with sparse arrays.')
            self.U = U
        else:
            self.dims = U.shape[:2]
            self.U = U.copy().reshape([-1,U.shape[-1]])
        if norm_svt:
            # Careful interpreting the results with normalized components...
            from sklearn.preprocessing import normalize
            normed = normalize(SVT,norm='l1',axis=1)
        # compute the covariance of the temporal components and estimate the pixelwise variance 
        self.cov_svt = np.cov(normed).astype('float32')
        self.var_pix = np.sum((self.U@self.cov_svt).T.conj()*self.U.T,axis = 0)
                
    def get(self,x,y):
        try:
            xy=np.ravel_multi_index([x,y],dims=self.dims)
        except ValueError:
            return None
        # Estimate the cov for a pixel from the covariance of the temporal components
        cov_pix = self.U[xy,:] @ self.cov_svt @ (self.U.T)
        # Normalize by the pixelwise
        std_pix = (self.var_pix[xy]**.5) * (self.var_pix**.5)
        cov_pix[xy] = np.nan
        return (cov_pix/std_pix).reshape(self.dims)

    def pylab_show(self):
        import pylab as plt
        fig = plt.figure()
        ax = fig.add_axes([.1,.1,.8,.8])
        self.pylab_im = ax.imshow(self.get(0,0),clim=[-1,1],cmap='RdBu_r')

        def update(event):
            if event.inaxes == ax:
                xy = [int(event.ydata),int(event.xdata)]
                self.pylab_im.set_data(self.get(*xy))
                fig.canvas.draw()
            fig.canvas.flush_events()
        fig.canvas.mpl_connect('button_press_event', update)
    

def _compute_percentile_from_reconstructed(U,SVT,percentile = 10):
    return np.percentile(np.dot(U,SVT), percentile, axis = 1)



#  wfield - tools to analyse widefield data - qt widgets 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *
from .allen import (load_allen_landmarks,
                    save_allen_landmarks,
                    allen_landmarks_to_image_space,
                    allen_transform_from_landmarks,
                    allen_load_reference,
                    allen_transform_regions)

from PyQt5.QtWidgets import (QApplication,
                             QWidget,
                             QTableWidgetItem,
                             QDockWidget,
                             QFormLayout,
                             QHBoxLayout,
                             QVBoxLayout,
                             QGridLayout,
                             QMainWindow,
                             QGroupBox,
                             QSlider,
                             QWidgetAction,
                             QListWidget,
                             QComboBox,
                             QPushButton,
                             QLabel,
                             QLineEdit,
                             QTextEdit,
                             QDoubleSpinBox,
                             QSpinBox,
                             QCheckBox,
                             QProgressBar,
                             QFileDialog,
                             QAction)

from PyQt5 import QtCore
from PyQt5.QtCore import Qt,QTimer

import pyqtgraph as pg
pg.setConfigOption('crashWarning', False)
pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'k')
pg.setConfigOptions(imageAxisOrder='row-major')
axiscolor = 'k'

class AllenMatchTable(QWidget):
    def __init__(self, landmarks_file = None,
                 reference = 'dorsal_cortex',
                 parent = None,
                 save_and_close = None):
        super(AllenMatchTable,self).__init__()
        # This widget shows: bregmaoffset in image coordinates, resolution, landmarks table
        self.reference = reference
        self.landmarks_file = landmarks_file
        if not os.path.isfile(landmarks_file):
            landmarks_file = None
        lmarks = load_allen_landmarks(landmarks_file,reference = reference)
        self.parent  = parent
        self.M = None
        if ('landmarks' in lmarks.keys() and 
            'bregma_offset' in lmarks.keys() and
            'resolution' in lmarks.keys()):
            landmarks_im = allen_landmarks_to_image_space(
                lmarks['landmarks'].copy(), 
                lmarks['bregma_offset'],
                lmarks['resolution'])
            x = landmarks_im['x']
            y = landmarks_im['y']
            if 'landmarks_match' in lmarks.keys():
                x = lmarks['landmarks_match']['x']
                y = lmarks['landmarks_match']['y']
            landmarks_im['nx'] = x
            landmarks_im['ny'] = y
        else:
            print('Need (landmarks,bregma_offset,resolution) in the landmarks file.')
            return
        self.table = pg.TableWidget()
        self.landmarks_im = landmarks_im
        self.landmarks = lmarks['landmarks']
        self.resolution = lmarks['resolution']
        self.bregma_offset = lmarks['bregma_offset']
        if 'transform' in lmarks.keys():
            self.M = lmarks['transform']
            self.parent.M = self.M
        self.table.setData(landmarks_im.to_records(index = False))
        lay = QFormLayout()
        self.setLayout(lay)

        w = QWidget()
        l = QFormLayout()
        w.setLayout(l)
        # resolution
        self.wres = QLineEdit()
        self.wres.setText(str(self.resolution))
        l.addRow(QLabel('resolution  [mm/pixel]'),self.wres)
        # bregma offset
        self.wbregma = QLineEdit()
        self.wbregma.setText('{0},{1}'.format(*self.bregma_offset))
        l.addRow(QLabel('bregma_offset'),self.wbregma)

        def ubregma():
            tt = self.wbregma.text()
            try:
                tt = tt.split(',')
                self.bregma_offset = [int(tt[0]),int(tt[1])]
            except Exception as e:
                print(e)
                print('bregma_offset: Needs 2 comma separated values.')
        self.wbregma.textChanged.connect(ubregma)

        def ures():
            tt = self.wres.text()
            try:
                self.resolution = float(tt)
            except Exception as e:
                print(e)
                print('Resolution needs to be a float.')
        self.wbregma.textChanged.connect(ubregma)
        # save transform
        if save_and_close is None:
            self.wsave = QPushButton('Save points')
        else:
            self.wsave = QPushButton('Save points and close')

        l.addRow(self.wsave)
        def usave():
            print('Saving points and landmarks to the _landmarks.json file.')
            self.usave()
            if not save_and_close is None:
                save_and_close.close()
        self.wsave.clicked.connect(usave)    
        lay.addRow(self.table,w)

    def usave(self):
        if self.parent.folder is None:
            print('Folder is not defined..')
        if self.landmarks_file is None: # then it is the dorsal_cortex
            fname = self.reference+'_landmarks.json'
        else:
            fname = os.path.basename(self.landmarks_file)
        fname = pjoin(self.parent.folder, fname)
        landmarks_match = self.landmarks.copy()
        landmarks_match.x = self.landmarks_im.nx
        landmarks_match.y = self.landmarks_im.ny
        landmarks_im = allen_landmarks_to_image_space(
            self.landmarks.copy(), 
            self.bregma_offset,
            self.resolution)
        self.M = allen_transform_from_landmarks(landmarks_im,landmarks_match)
        save_allen_landmarks(self.landmarks,filename = fname,
                             resolution = self.resolution,
                             bregma_offset = self.bregma_offset,
                             landmarks_match = landmarks_match,
                             transform = self.M)
        self.parent.M = self.M

class AllenArea(pg.PlotCurveItem):
    def __init__(self,area, side, *args,**kwargs):
        super().__init__(*args,**kwargs)
        self.area = area
        self.side = side
        x = area[side+'_x']
        y = area[side+'_y']
        self.setData(x=x, y=y)
        self.setAcceptHoverEvents(False)
    def highlight(self,color = None):
        if color is None:
            color = self.area['allen_rgb']
        self.setPen(pg.mkPen(color=color,
                             width=5))
        
    def hoverEnterEvent(self, ev):
        self.highlight()
        self.text = pg.TextItem(text = self.area['acronym'])
        self.text.setPos(*self.area[self.side+'_center'][::-1])
        self.scene().addItem(self.text)
    def hoverLeaveEvent(self, ev):
        self.setPen(pg.mkPen(color='w',size=1))
        self.scene().removeItem(self.text)
        
class QAllenAreasPlot():
    def __init__(self,plot,
                 reference='dorsal_cortex',
                 sides = ['left','right'],
                 parent = None,
                 resolution = 0.01,
                 bregma_offset = [0,0]):
        self.ccf_regions,self.proj,self.brain_outline = allen_load_reference(reference)
        self.parent = parent
        self.landmarks = None
        self.resolution = resolution
        self.sides = sides
        self._parameters_from_gui()
        self.plot = plot
        self.plot_items = []
        self.plot_text = []
        self.plot_centers = []
        self.warped_image = False

    def _parameters_from_gui(self):
        if hasattr(self.parent,'parent'):
            if hasattr(self.parent.parent,'allenparwidget'):
                self.landmarks = self.parent.parent.allenparwidget.landmarks
                self.resolution = self.parent.parent.allenparwidget.resolution
                self.bregma_offset = self.parent.parent.allenparwidget.bregma_offset
                self.M = self.parent.parent.allenparwidget.M
            if hasattr(self.parent,'warp_im'):
                self.warped_image = self.parent.warp_im 

    def highlight(self,idx,side,color = None):
        s = self.sides.index(side)
        idx *= 2
        idx += s
        self.plot_items[int(idx)].highlight(color)
        
    def update(self):
        self._parameters_from_gui()
        if self.warped_image:
            M = None
        else:
            M = self.M
        self.nccf_regions = allen_transform_regions(M, self.ccf_regions,
                                                    resolution = self.resolution,
                                                    bregma_offset = self.bregma_offset)
        self.remove()
        for iarea,area in self.nccf_regions.iterrows():
            for side in self.sides:
                self.plot_items.append(AllenArea(area,side))
                self.plot_text.append(area['acronym'])
                self.plot_centers.append(area[side+'_center'])
                self.plot.addItem(self.plot_items[-1])
                self.plot_items[-1].setClickable(True, width=2)
                
    def remove(self):
        for p in self.plot_items:
            p.setVisible(False)
            p.scene().removeItem(p)
        self.plot_items = []
        self.plot_text = []
        self.plot_centers = []

        
class ImageWidget(QWidget):
    def __init__(self):
        super(ImageWidget,self).__init__()

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        self.win = pg.GraphicsLayoutWidget()
        self.layout.addWidget(self.win)

        self.pl = self.win.addPlot()
        self.pl.getViewBox().invertY(True)
        self.pl.getViewBox().setAspectLocked(True)
        
        self.im = pg.ImageItem()
        self.win.setCentralWidget(self.pl)
        self.pl.addItem(self.im)
        
        self.pl.getAxis('bottom').setPen(axiscolor)
        self.pl.getAxis('left').setPen(axiscolor)
        self.pl.setClipToView(True)
        
    def _add_hist(self):
        self.hist = pg.HistogramLUTItem()
        self.hist.setImageItem(self.im)
        self.pl.addItem(self.hist)
            
class CustomDragPoints(pg.GraphItem):
    def __init__(self):
        self.drag_points = None
        self.drag_offset = None
        self.text = []
        self.text_items = []
        pg.GraphItem.__init__(self)

    def setData(self, points = None, text = None):
        # pass a list of dictionaries (with pos and text as keys)
        if points is None:
            return
        if text is None:
            text = self.text
        else:
            self.text = text
        self.points = points
        self.scatter.setData(self.points)
        for i in self.text_items:
            i.scene().removeItem(i)
        self.text_items = []
        for p,t in zip(self.points,text):
            self.text_items.append(pg.TextItem(text=t,color='y'))
            self.text_items[-1].setParentItem(self)
            self.text_items[-1].setPos(*p['pos'])
        self.informViewBoundsChanged()
    def set_visible(self,val):
        self.scatter.setVisible(val)
        for t in self.text_items:
            t.setVisible(val)
            
    def _findind(self,pt):
        for i,p in enumerate(self.points):
            if ((pt.pos().x() == p['pos'][0]) and
                (pt.pos().y() == p['pos'][1])):
                ind = i
        return ind
        
    def mouseDragEvent(self, ev):
        if ev.button() != QtCore.Qt.LeftButton:
            ev.ignore()
            return
        if ev.isStart():
            pos = ev.buttonDownPos()
            pts = self.scatter.pointsAt(pos)
            if len(pts) == 0:
                ev.ignore()
                return
            self.dragPoint = self._findind(pts[0])
            self.dragOffset = self.points[self.dragPoint]['pos'] - pos
        elif ev.isFinish():
            self.dragPoint = None
            return
        else:
            if self.dragPoint is None:
                ev.ignore()
                return
        self.points[self.dragPoint]['pos'] = ev.pos() + self.dragOffset
        self.setData(points = self.points)
        ev.accept()
        
        #ind = self._findind()

        
class RawDisplayWidget(ImageWidget):
    def __init__(self, stack,parent=None,pointsize = 10,
                 reference = 'dorsal_cortex'):
        super(RawDisplayWidget,self).__init__()
        self.parent = parent
        self.stack = stack
        self.referencename = reference
        if hasattr(self.parent,'roiwidget'):
            self.roiwidget = self.parent.roiwidget
        self.regions_plot = []
        self.iframe = np.clip(100,0,len(self.stack))
        self.ichan  = 0
        self.warp_im = False
        tmp = self.stack[:np.clip(100,0,len(self.stack))]
        self._init_ui()
        self._add_hist()
        self.adaptative_histogram = False
        self.allen_show_areas = False
        self.set_image(self.iframe)
        self.hist.setLevels(*np.percentile(tmp,[1,99]))

        self.win.scene().sigMouseClicked.connect(self.mouseMoved) # not ready yet    
        
        if hasattr(self.parent,'allenparwidget'):
            self.allenwidget = self.parent.allenparwidget
            self.points = CustomDragPoints()
            tmp = []
            tmp_text = []
            for i,t in self.allenwidget.landmarks_im.iterrows():
                tmp.append(dict(pos = (t.nx,t.ny),
                                size = pointsize,
                                pen=pg.mkPen(color=t.color)))
                tmp_text.append(t['name'])
            self.points.setData(tmp,tmp_text)
            self.pl.addItem(self.points)
            def update_table():
                # Update the table when points move
                try:
                    nx = [p['pos'][0] for p in self.points.points]
                    ny = [p['pos'][1] for p in self.points.points]
                    self.allenwidget.landmarks_im.nx = nx
                    self.allenwidget.landmarks_im.ny = ny
                    for i,(x,y) in enumerate(zip(nx,ny)):
                        self.allenwidget.table.item(i,4).setText(str(x))
                        self.allenwidget.table.item(i,5).setText(str(y))
                    try:
                        self.allenwidget.usave()
                    except Exception as e:
                        print(e)
                    if self.wallen.isChecked():
                        self.allenplot.update()
                    else:
                        self.allenplot.remove()
                except:
                    print('There was an error updating the points.')
            self.points.scatter.sigPlotChanged.connect(update_table)
            self.allenplot = QAllenAreasPlot(plot=self.pl,parent = self,
                                             reference = self.referencename)

    def _init_ui(self):
        widget = QWidget()
        slayout = QFormLayout()
        widget.setLayout(slayout)
        self.wframe = QSlider(Qt.Horizontal)
        self.wframe.setValue(0)
        self.wframe.setMaximum(len(self.stack)-1)
        self.wframe.setMinimum(0)
        self.wframe.setSingleStep(1)
        self.wframelabel = QLabel('Frame {0:d}:'.format(self.wframe.value()))
        slayout.addRow(self.wframelabel, self.wframe)
        self.wchan = QSlider(Qt.Horizontal)
        self.wchan.setValue(0)
        self.wchan.setMaximum(self.stack.shape[1]-1)
        self.wchan.setMinimum(0)
        self.wchan.setSingleStep(1)
        self.wchanlabel = QLabel('Channel {0:d}:'.format(self.wchan.value()))
        w1 = QWidget()
        l1 = QHBoxLayout()
        w1.setLayout(l1)
        l1.addWidget(self.wchanlabel)
        l1.addWidget(self.wchan)
        self.wimadapt = QCheckBox()
        l1.addWidget(QLabel('Highlight vessels:'))
        l1.addWidget(self.wimadapt)
        allenframe = QGroupBox(self)
        allenframe.setTitle('Allen Common Coordinates Framework')
        allenl = QFormLayout(allenframe)
        self.wimwarp = QCheckBox()
        allenl.addRow(QLabel('Show transformed frames:'),self.wimwarp)
        self.wallen = QCheckBox()
        allenl.addRow(QLabel('Plot areas:'),self.wallen)
        l1.addWidget(allenframe)
        slayout.addRow(w1)
        self.layout.addWidget(widget)
        
        def uhist(val):
            self.adaptative_histogram = val
            self.set_image()
        def uwarp(val):
            self.warp_im = val
            self.set_image()
            self.points.set_visible(not val)
            if self.allen_show_areas:
                self.allenplot.update()
        def uallen(val):
            self.allen_show_areas = val
            if val:
                self.allenplot.update()
            else:
                self.allenplot.remove()
        def uframe(val):
            i = self.wframe.value()
            self.wframelabel.setText('Frame {0:d}:'.format(i))
            self.set_image(i)

        def uchan(val):
            self.ichan = self.wchan.value()
            self.wchanlabel.setText('Channel {0:d}:'.format(self.ichan))
            self.set_image()
        
        self.wframe.valueChanged.connect(uframe)
        self.wimadapt.stateChanged.connect(uhist)
        self.wimwarp.stateChanged.connect(uwarp)
        self.wchan.valueChanged.connect(uchan)
        self.wallen.stateChanged.connect(uallen)

    def set_image(self,i=None,):
        if not i is None:
            self.iframe = i
        img = self.stack[np.clip(self.iframe,0,self.stack.shape[0]-1),self.ichan]
        if self.adaptative_histogram:
            img = im_adapt_hist(img)
        if self.warp_im:
            if hasattr(self.parent,'M'):
                img = im_apply_transform(img.squeeze(),self.parent.M)
        self.im.setImage(img.squeeze(), autoLevels=False)
        if hasattr(self,'roiwidget'):
            self.roiwidget.line.setPos((self.iframe,0))
            self.roiwidget.update()
        
    def on_roi_update(self,i):
        idx = self.roiwidget.get_roi_flatidx(i)
        xidx = self.roiwidget.xidx + self.iframe
        xidx = np.clip(xidx,0,len(self.stack))
        s = self.stack[xidx[0]:xidx[1],self.ichan]
        s = s.reshape([s.shape[0],-1])
        t = np.nanmean(s[:,idx],
                       axis=1).astype('float32')
        time = np.arange(xidx[0],xidx[1])
        self.roiwidget.plots[i].setData(x = time,
                                        y = t+self.roiwidget.offset*i)#,connect=self.trial_mask)

    def mouseMoved(self,pos):
        modifiers = QApplication.keyboardModifiers()
        pos = self.im.mapFromScene(pos.scenePos())
        if (bool(modifiers == Qt.ControlModifier) and
            hasattr(self,'roiwidget')):            
            updatefunc = partial(self.on_roi_update, i=len(self.roiwidget.rois))
            self.parent.roiwidget.add_roi((pos.x(),pos.y()),
                                          roitarget = self.im,
                                          roiscene=self.pl,
                                          ROI = True,
                                          updatefunc = updatefunc)
        
        
class SVDDisplayWidget(ImageWidget):
    def __init__(self, stack,
                 parent=None,reference = 'dorsal_cortex'):
        super(SVDDisplayWidget,self).__init__()
        self.parent = parent
        self.stack = stack
        self.warp_im = stack.warped
        self.referencename = reference
        self.roiwidget = self.parent.roiwidget
        self.regions_plot = []
        self.iframe = np.clip(51,1,len(self.stack))
        tmp = self.stack[1:np.clip(100,0,len(self.stack))]
        self._init_ui()
        self._add_hist()
        self.allen_show_areas = False
        self.set_image(self.iframe)
        self.win.scene().sigMouseClicked.connect(self.mouseMoved)    
        self.hist.setLevels(*np.percentile(tmp,[1,99]))

    def _init_ui(self):
        w = QWidget()
        l = QHBoxLayout()
        w.setLayout(l)
        self.wframe = QSlider(Qt.Horizontal)
        self.wframe.setValue(0)
        self.wframe.setMaximum(len(self.stack)-1)
        self.wframe.setMinimum(0)
        self.wframe.setSingleStep(1)
        self.wframelabel = QLabel('Frame {0:d}:'.format(self.wframe.value()))
        l.addWidget(self.wframelabel)
        l.addWidget(self.wframe)

        allenframe = QGroupBox(self)
        allenframe.setTitle('Allen Common Coordinates Framework')
        allenl = QFormLayout(allenframe)
        self.wimwarp = QCheckBox()
        self.wimwarp.setChecked(self.warp_im)
        allenl.addRow(QLabel('Show transformed frames:'),self.wimwarp)
        self.wallen = QCheckBox()
        allenl.addRow(QLabel('Plot areas:'),self.wallen)
        l.addWidget(allenframe)
        self.layout.addWidget(w)
        def uwarp(val):
            self.warp_im = val
            if hasattr(self.parent,'M'):
                self.stack.set_warped(val,self.parent.M)
            if self.allen_show_areas:
                self.allenplot.update()
            self.set_image()
        def uallen(val):
            self.allen_show_areas = val
            if val:
                if not hasattr(self,'allenplot'):
                    self.allenplot = QAllenAreasPlot(plot=self.pl,parent = self,
                                                     reference=self.referencename)
                self.allenplot.update()
            else:
                self.allenplot.remove()
                del self.allenplot

        def uframe(val):
            i = self.wframe.value()
            self.wframelabel.setText('Frame {0:d}:'.format(i))
            self.set_image(i)
            self.roiwidget.line.setPos((i,0))
            self.roiwidget.update()
        self.wframe.valueChanged.connect(uframe)
        self.wimwarp.stateChanged.connect(uwarp)
        self.wallen.stateChanged.connect(uallen)

    def set_image(self,i=None):
        if not i is None:
            self.iframe = i
        img = self.stack[self.iframe]
        self.im.setImage(img.squeeze(), autoLevels = False)
        
    def get_xy(self,x,y):
        x = int(np.clip(x,0,self.stack.shape[1]))
        y = int(np.clip(y,0,self.stack.shape[2]))
        idx = np.ravel_multi_index((x,y),self.stack.shape[1:])
        t = np.asarray(self.stack.Uflat[idx,:].dot(self.stack.SVT))
        return t

    def on_roi_update(self,i):
        idx = self.roiwidget.get_roi_flatidx(i)
        xidx = self.roiwidget.xidx + self.iframe
        xidx = np.clip(xidx,0,len(self.stack))
        if issparse(self.stack.Uflat):
            t = np.nanmean(np.asarray(self.stack.Uflat[idx,:].todense()) @ self.stack.SVT[:,xidx[0]:xidx[1]],
                           axis=0).astype('float32')

        else:
            t = np.nanmean(self.stack.Uflat[idx,:] @ self.stack.SVT[:,xidx[0]:xidx[1]],
                           axis=0).astype('float32')
        time = np.arange(xidx[0],xidx[1])
        self.roiwidget.plots[i].setData(x = time,
                                        y = t+self.roiwidget.offset*i)#,connect=self.trial_mask)

    def plot_allenroi(self,i,idx):
        if issparse(self.stack.Uflat):
            t = np.nanmean(np.asarray(self.stack.Uflat[idx,:].todense()) @ self.stack.SVT,
                           axis=0).astype('float32')
        else:
            t = np.nanmean(self.stack.Uflat[idx,:] @ self.stack.SVT,
                           axis=0).astype('float32')
        time = np.arange(0,t.shape[0])
        self.roiwidget.plots[i].setData(x = time,
                                        y = t+self.roiwidget.offset*i)#,connect=self.trial_mask)
    def on_allenroi_update(self):
        self.roiwidget.p1.setRange(xRange=self.roiwidget.xidx+self.iframe)
        
    def mouseMoved(self,pos):
        modifiers = QApplication.keyboardModifiers()
        pos = self.im.mapFromScene(pos.scenePos())
        if bool(modifiers == Qt.ControlModifier):            
            updatefunc = partial(self.on_roi_update, i=len(self.roiwidget.rois))
            self.parent.roiwidget.add_roi((pos.x(),pos.y()),
                                          roitarget = self.im,
                                          roiscene=self.pl,
                                          ROI = True,
                                          updatefunc = updatefunc)
        else:
            if hasattr(self,'allenplot'):
                p = (pos.x(),pos.y())
                region,side,i = point_find_ccf_region(p,self.allenplot.nccf_regions)
                if not region is None:
                    if not [i,side] in self.regions_plot:
                        self.regions_plot.append([i,side])
                        H = contour_to_mask(region[side + '_x'],
                                            region[side + '_y'],
                                            self.stack.shape[1:])
                        idx = np.ravel_multi_index(np.where(H==1),self.stack.shape[1:])
                        color = self.parent.roiwidget.add_roi(
                            None,
                            roitarget = self.im,
                            roiscene=self.pl,
                            ROI = False,
                            updatefunc = self.on_allenroi_update)

                        self.plot_allenroi(len(self.roiwidget.rois)-1,idx)
                        color = tuple(int(color.strip('#')[i:i+2], 16) for i in (0, 2, 4))
                        self.allenplot.highlight(i,side,color)

class SVDViewer(QMainWindow):
    def __init__(self,stack, folder = None, raw = None, reference = 'dorsal_cortex',
                 trial_onsets = None,
                 start_correlation = False):
        super(SVDViewer,self).__init__()
        self.setWindowTitle('wfield')
        self.folder = folder
        if self.folder is None:
            self.folder = os.path.abspath(os.path.curdir)
        self.referencename = reference
        landmarks_file = pjoin(self.folder,reference+'_landmarks.json')

        self.stack = stack
        self.trial_onsets = trial_onsets
        self.raw = raw
        self.trial_mask = np.ones((self.stack.SVT.shape[-1]),dtype=bool)
        if not self.trial_onsets is None:
            self.trial_mask[self.trial_onsets[:,1]] = False
        self.setDockOptions(QMainWindow.AllowTabbedDocks |
                            QMainWindow.AllowNestedDocks |
                            QMainWindow.AnimatedDocks)
        self.roiwidget = ROIPlotWidget(stack)        
        self.displaywidget = SVDDisplayWidget( self.stack,parent = self,reference = self.referencename)
        if start_correlation:
            self.localcorrwidget = LocalCorrelationWidget(stack)
            
        if not self.raw is None:
            self.allenparwidget = AllenMatchTable(landmarks_file = landmarks_file,
                                                  reference = self.referencename,
                                                  parent = self) 
            self.rawwidget = RawDisplayWidget(raw,
                                              parent = self,
                                              reference = self.referencename)
            
        self.svdtab = QDockWidget('Reconstructed')
        self.svdtab.setWidget(self.displaywidget)
        self.addDockWidget(Qt.TopDockWidgetArea,self.svdtab)        
        # Raw data
        if not raw is None:
            self.rawtab = QDockWidget('Raw data')
            self.rawtab.setWidget(self.rawwidget)
            self.addDockWidget(Qt.TopDockWidgetArea,self.rawtab)
        # Pixel correlation 
        if hasattr(self,'localcorrwidget'):
            self.lcorrtab = QDockWidget('Pixel correlation')
            self.lcorrtab.setWidget(self.localcorrwidget)
            self.addDockWidget(Qt.TopDockWidgetArea,self.lcorrtab)
        if not raw is None:
            if hasattr(self,'lcorrtab'):
                self.tabifyDockWidget(self.lcorrtab,self.rawtab)
            self.tabifyDockWidget(self.rawtab,self.svdtab)
        else:
            if hasattr(self,'lcorrtab'):
                self.tabifyDockWidget(self.lcorrtab,self.svdtab)
        self.roitab = QDockWidget("ROI", self)
        self.roitab.setWidget(self.roiwidget)
        self.addDockWidget(Qt.BottomDockWidgetArea,self.roitab)
        self.set_dock(self.roitab,False)
        # Allen match
        if hasattr(self,'allenparwidget'):
            self.allenpartab = QDockWidget('CCF match parameters')
            self.allenpartab.setWidget(self.allenparwidget)
            self.addDockWidget(Qt.BottomDockWidgetArea,self.allenpartab)
            self.tabifyDockWidget(self.allenpartab,self.roitab)        
        self.show()

    def set_dock(self,dock,floating=False):
        dock.setAllowedAreas(Qt.LeftDockWidgetArea |
                             Qt.RightDockWidgetArea |
                             Qt.BottomDockWidgetArea |
                             Qt.TopDockWidgetArea)
        dock.setFeatures(QDockWidget.DockWidgetMovable |
                         QDockWidget.DockWidgetFloatable)
        dock.setFloating(floating)

class RawViewer(QMainWindow):
    def __init__(self,raw,
                 folder = None,
                 reference = 'dorsal_cortex',
                 trial_onsets = None):
        super(RawViewer,self).__init__()
        self.setWindowTitle('wfield')
        self.raw = raw
        self.folder = folder
        if self.folder is None:
            self.folder = os.path.abspath(os.path.curdir)
        self.referencename = reference
        landmarks_file = pjoin(self.folder,reference+'_landmarks.json')
        
        self.trial_onsets = trial_onsets
        self.trial_mask = np.ones((self.raw.shape[0]),dtype=bool)
        if not self.trial_onsets is None:
            self.trial_mask[self.trial_onsets[:,1]] = False
        self.setDockOptions(QMainWindow.AllowTabbedDocks |
                            QMainWindow.AllowNestedDocks |
                            QMainWindow.AnimatedDocks)
        self.roiwidget = ROIPlotWidget(self.raw)  
        if not self.raw is None:
            self.allenparwidget = AllenMatchTable(landmarks_file = landmarks_file,
                                                  reference = self.referencename,
                                                  parent = self)
            self.rawwidget = RawDisplayWidget(raw,
                                              parent = self,
                                              reference = self.referencename)
            
        # Raw data
        if not raw is None:
            self.rawtab = QDockWidget('Raw data')
            self.rawtab.setWidget(self.rawwidget)
            self.addDockWidget(Qt.TopDockWidgetArea,self.rawtab)
            
        self.roitab = QDockWidget("ROI", self)
        self.roitab.setWidget(self.roiwidget)
        self.addDockWidget(Qt.BottomDockWidgetArea,self.roitab)
        self.set_dock(self.roitab,False)
        # Allen match
        if hasattr(self,'allenparwidget'):
            self.allenpartab = QDockWidget('CCF match parameters')
            self.allenpartab.setWidget(self.allenparwidget)
            self.addDockWidget(Qt.BottomDockWidgetArea,self.allenpartab)
            if hasattr(self,'roitab'):
                self.tabifyDockWidget(self.roitab,self.allenpartab)        
        self.show()

    def set_dock(self,dock,floating=False):
        dock.setAllowedAreas(Qt.LeftDockWidgetArea |
                             Qt.RightDockWidgetArea |
                             Qt.BottomDockWidgetArea |
                             Qt.TopDockWidgetArea)
        dock.setFeatures(QDockWidget.DockWidgetMovable |
                         QDockWidget.DockWidgetFloatable)
        dock.setFloating(floating)

class AllenMatchWidget(QWidget):
    def __init__(self,raw,
                 folder = None,
                 reference = 'dorsal_cortex',
                 mainwidget = None):
        super(AllenMatchWidget,self).__init__()
        self.raw = raw
        self.folder = folder
        if self.folder is None:
            self.folder = os.path.abspath(os.path.curdir)
        self.referencename = reference
        landmarks_file = pjoin(self.folder,reference+'_landmarks.json')
        
        self.allenparwidget = AllenMatchTable(landmarks_file = landmarks_file,
                                              reference = self.referencename,
                                              parent = self,
                                              save_and_close = mainwidget) # to close with the save button...
        self.rawwidget = RawDisplayWidget(raw,
                                          parent = self,
                                          reference = self.referencename)
        l = QHBoxLayout()
        self.setLayout(l)
        l.addWidget(self.rawwidget)
        l.addWidget(self.allenparwidget)
        self.show()

        
class LocalCorrelationWidget(ImageWidget):
    def __init__(self, stack,parent=None):
        super(LocalCorrelationWidget,self).__init__()
        self.parent = parent
        self.stack = stack
        U = stack.U
        SVT = stack.SVT
        from .utils_svd import svd_pix_correlation
        try:
            self.localcorr = svd_pix_correlation(U,SVT,
                                                 dims = stack.shape[1:],
                                                 norm_svt=True)
        except:
            print('Could not compute correlation (sparse mode?)')
            return
        self.levels = [0,1]
        
        pos = np.array([0, 0.5, 1.])
        color = np.array([[0,0,255,255], [255,255,255,255], [255,0,0,255]], dtype=np.ubyte)
        cmap = pg.ColorMap(pos, color)
        lut = cmap.getLookupTable(0, 1., 256)

        self._add_hist()
        self.hist.gradient.setColorMap(cmap)
        self.hist.setLevels(*self.levels)

        self.set_image([100,100])
        self.win.scene().sigMouseMoved.connect(self.mouseMoved)
        
    def set_image(self,xy=[0,0]):
        img = self.localcorr.get(*xy)
        if not img is None:
            img += 1.
            img /= 2.
        self.im.setImage(img.squeeze(),autoLevels = False)

    def mouseMoved(self,pos):
        modifiers = QApplication.keyboardModifiers()
        pos = self.im.mapFromScene(pos)
        if bool(modifiers == Qt.ControlModifier):
            self.set_image((int(pos.y()),int(pos.x())))

class ROIPlotWidget(QWidget):
    colors = ['#d62728',
          '#1f77b4',
          '#ff7f0e',
          '#2ca02c',
          '#9467bd',
          '#8c564b',
          '#e377c2',
          '#7f7f7f',
          '#bcbd22']
    penwidth = 1

    def __init__(self,stack,npoints = 500):
        super(ROIPlotWidget,self).__init__()	
        layout = QGridLayout()
        self.stack = stack
        self.setLayout(layout)
        win = pg.GraphicsLayoutWidget(parent=self)
        self.p1 = win.addPlot()
        self.p1.getAxis('bottom').setPen(axiscolor) 
        self.p1.getAxis('left').setPen(axiscolor) 
        layout.addWidget(win,0,0)
        self.line = pg.InfiniteLine(angle = 90,movable = True)
        self.p1.addItem(self.line)
        self.xidx = np.array([-npoints,npoints])
        self.N = npoints
        self.rois = []
        self.rois_parent = []
        self.rois_scene = []
        self.plot_updates = []
        self.plots = []
        self.buffers = []
        self.time = np.arange(self.stack.shape[0],dtype='float32')
        self.offset = 0.1
        
    def add_roi(self,pos,roitarget = None, roiscene = None, ROI=True,updatefunc = None):
        pencolor = self.colors[
            np.mod(len(self.plots),len(self.colors))]
        if ROI:
            self.rois.append(pg.RectROI(pos=pos,
                                        size=20,
                                        pen=pencolor))
            roiscene.addItem(self.rois[-1])
            self.rois[-1].sigRegionChanged.connect(updatefunc)
            self.rois_parent.append(roitarget)
            self.rois_scene.append(roiscene)
        else:
            self.rois.append(None)
            self.rois_parent.append(None)
            self.rois_scene.append(None)
        self.plots.append(pg.PlotCurveItem(pen=pg.mkPen(
            color=pencolor,width=self.penwidth)))
        self.p1.addItem(self.plots[-1])
        self.plot_updates.append(updatefunc)
        updatefunc()
        return pencolor
    
    def items(self):
        return self.rois
    def update(self):
        for p in self.plot_updates:
            p()
    def closeEvent(self,ev):
        for i,roi in enumerate(self.rois):
            self.rois_scene[i].removeItem(roi)
        ev.accept()
        
    def get_roi_flatidx(self,i):
        X = np.zeros(self.stack.shape[-2:],dtype='uint8')
        r = self.rois[i].getArraySlice(X, self.rois_parent[i], axes=(0,1))
        X[r[0][0],r[0][1]]=1
        return np.ravel_multi_index(np.where(X==1),self.stack.shape[-2:])






from .utils import *
from numpy.fft import fft
import math
from scipy.ndimage.filters import gaussian_filter

def mask_to_3d(mask,shape,include_mask = None):
    '''
    3dmask = mask_to_3d(mask,shape,include_mask = None)
    
    Broadcasts a mask to a 3d array.

    Usage:
s1mask = mask_contour_on_extent(nrefregions[2]['c'],oriim.shape,extent=extent)[:,::-1]
v1mask = mask_contour_on_extent(nrefregions[6]['c'],oriim.shape,extent=extent)[:,::-1]
rlmask = mask_contour_on_extent(nrefregions[10]['c'],oriim.shape,extent=extent)[:,::-1]
s1moviemask = mask_to_3d([v1mask,s1mask,rlmask],tmp.shape,include_mask = winmask)
'''
    if type(mask) is list:
        mask = np.any(np.stack([m.astype(bool) for m in mask]),axis = 0)
    nmask = mask.copy().astype(mask.dtype)
    if not include_mask is None:
        nmask = np.all([nmask,include_mask],axis = 0)
    return np.broadcast_to(nmask,shape)

def get_signals_from_mask(dat, mask,include_mask = None):
    '''
    sigs = get_signals_from_mask(dat, mask,include_mask = None)

    Extract signals from a 2d or 3d array.
    
    Usage:
s1res = get_signals_from_mask(stimavgsf[0],s1mask,include_mask = winmask)
'''
    if type(mask) is list:
        mask = np.any(np.stack([m.astype(bool) for m in mask]),axis = 0)
    if not len(mask.shape) == len(dat.shape):
        mask = mask_to_3d(mask,dat.shape,include_mask = include_mask)
    res = dat[mask]
    if len(dat.shape) > 2:
        res = res.reshape((dat.shape[0],int(res.shape[0]/dat.shape[0])))
    return res

#############################################################
############Pseudo color and phase analysis##################
#############################################################

def hsv_colorbar_image():
    '''
    hsv_colorbar_image()
    
    
    Dumb way to make a colorbar:for an hsv image
    cbarim = hsv_colorbar_image()
    plt.imshow(cbarim,aspect= 'auto',extent = [0,1,-40,40])
    
    '''
    hsvimg = np.ones([1,255,3])*255
    hsvimg[:,:,0] = np.linspace(0,255,255)
    hsvimg = hsvimg.transpose([1,0,2])
    return cv2.cvtColor(hsvimg.astype(np.uint8), cv2.COLOR_HSV2RGB_FULL)

def fft_movie(movie, component = 1,output_raw = False):
    '''
    Computes the fft of a movie and returns the magnitude and phase 
    '''
    movief = fft(movie, axis = 0)
    if output_raw:
        return movief[component]
    phase  = -1. * np.angle(movief[component]) % (2*np.pi)
    mag = (np.abs(movief[component])*2.)/len(movie)
    return mag,phase

def fft_get_phase(movief):
    return -1. * np.angle(movief) % (2*np.pi)

def gaussfilt2d(X,size = int(7)):
    '''
    2d gaussian filter a movie of images in parallel
    '''
    return np.stack(runpar(gaussian_filter,X,sigma = size))


def phasemap_to_visual_degrees(phasemap,startdeg,stopdeg):
    '''
    Normalizes the phasemap to visual angles
    Joao Couto 2019
    '''
    res = phasemap.copy() - np.nanmin(phasemap)
    res /= np.nanmax(res)
    res *= np.abs(np.diff([startdeg,stopdeg]))
    res += startdeg
    return res

def visual_sign_map(phasemap1, phasemap2):
    '''
    Computes the visual sign map from azimuth and elevation phase maps
    This is adapted from the Allen retinotopy code
    '''
    gradmap1 = np.gradient(phasemap1)
    gradmap2 = np.gradient(phasemap2)
    import scipy.ndimage as ni
    graddir1 = np.zeros(np.shape(gradmap1[0]))
    graddir2 = np.zeros(np.shape(gradmap2[0]))
    for i in range(phasemap1.shape[0]):
        for j in range(phasemap2.shape[1]):
            graddir1[i, j] = math.atan2(gradmap1[1][i, j], gradmap1[0][i, j])
            graddir2[i, j] = math.atan2(gradmap2[1][i, j], gradmap2[0][i, j])
    vdiff = np.multiply(np.exp(1j * graddir1), np.exp(-1j * graddir2))
    areamap = np.sin(np.angle(vdiff))
    return areamap

def im_fftphase_hsv(mov,component = 1,blur = 0,vperc=98,sperc=90,return_hsv=False):
    '''
    im_fftphase_hsv(mov,blur = 0,vperc=99,sperc=90)
    
    Creates a color image colorcoding the frame with fourier phase for each pixel 
    
        mov can be a 3d array or a list with the [magnitude, phase]
'''
    if not blur == 0:
        mov = runpar(im_gaussian,mov,sigma=blur)
    if not type(mov) is list:
        mag,H = fft_movie(mov,component = component)
    else:
        mag,H = mov
    H = H/(2*np.pi)
    V = mag.copy()
    V /= np.percentile(mag,vperc)
    S = mag**0.3
    S /= np.percentile(S,sperc)
    if return_hsv:
        return np.stack([H,S,V],axis=2).astype(np.float32)
    # Normalization for opencv ranges 0-255 for uint8
    hsvimg = np.clip(np.stack([H,S,V],axis=2).astype(np.float32),0,1)
    hsvimg *= 255
    return cv2.cvtColor(hsvimg.astype(np.uint8), cv2.COLOR_HSV2RGB_FULL)

def im_argmax_hsv(mov,blur = 0,vperc=98,sperc=90,return_hsv=False):
    '''
    im_argmax_hsv(mov,blur = 0,vperc=99,sperc=90)
    
    Creates a color image colorcoding the frame with largest amplitude for each pixel 
    '''
    if not blur == 0:
        mov = runpar(im_gaussian,mov,sigma=blur)
    H = np.argmax(mov,axis = 0)/np.float32(len(mov))
    maxim = np.max(mov,axis=0)
    meanim = np.mean(mov,axis=0)
    V = maxim.copy()
    V /= np.percentile(maxim,vperc)
    S = maxim-meanim
    S /= np.percentile(S,sperc)
    if return_hsv:
        return np.stack([H,S,V],axis=2).astype(np.float32)
    # Normalization for opencv ranges 0-255 for uint8
    hsvimg = np.clip(np.stack([H,S,V],axis=2).astype(np.float32),0,1)
    hsvimg *= 255
    return cv2.cvtColor(hsvimg.astype(np.uint8), cv2.COLOR_HSV2RGB_FULL)

def im_combineproj_hsv(stacks,
                       proj_funct = lambda x: np.nanstd(x,axis = 0),
                       vperc=98, sperc=98,
                       return_hsv=False):
    '''
    im_combineproj_hsv (stacks,
                        proj_funct=lambda x: np.std(x,axis = 0),
                        vperc=99, sperc=98)
    
    Creates a color image colorcoding the projections in space 
    '''
    proj = np.stack([proj_funct(x) for x in stacks])
    V,H = fft_movie(proj)
    H = H/(2*np.pi)
    S = V.copy()
    S = S
    S /= np.percentile(S,sperc)
    V /= np.percentile(V,vperc)
    if return_hsv:
        return np.stack([H,S,V],axis=2).astype(np.float32)
    # Normalization for opencv ranges 0-255 for uint8 
    hsvimg = np.clip(np.stack([H,S,V],axis=2),0,1)
    hsvimg[:,:,1:] *= 255
    hsvimg[:,:,0] *= 180
    return cv2.cvtColor(hsvimg.astype(np.uint8),
                        cv2.COLOR_HSV2RGB)                   




#  wfield - tools to analyse widefield data - allen tools 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *
from skimage import measure
from skimage.filters import gaussian
from scipy import ndimage
import json

# allensdk imports are inside the functions because it needs an old pandas and i dont understand why. To install it do: pip install allensdk

annotation_dir = wfield_dir #pjoin(os.path.expanduser('~'),'.wfield')

selection_dorsal_cortex = [
    'MOB',
    'FRP',
    'MOp',
    'MOs',
    'SSp-n',
    'SSp-m',
    'SSp-un',
    'PL',
    'ACAd',
    'RSPv',
    'RSPd',
    'RSPagl',
    'VISC',
    'SSs',
    'SSp-bfd',
    'SSp-tr',
    'SSp-ll',
    'SSp-ul',
    'TEa',    
    'AUDd',
    'AUDp',
    'AUDpo',
    'AUDv',
    'VISli',
    'VISpor',
    'VISpl',
    'VISpm',
    'VISl',
    'VISal',
    'VISrl',
    'VISa',
    'VISam',
    'VISp',]

selection_vis_som = ['RSPv',
                     'RSPd',
                     'RSPagl',
                     'SSs',
                     'SSp-bfd',
                     'SSp-tr',
                     'SSp-ll',
                     'SSp-ul',
                     'SSp-un',
                     'SSp-n',
                     'VISli',
                     'VISpor',
                     'VISp',
                     'VISpm',
                     'VISl',
                     'VISal',
                     'VISrl',
                     'VISpl',
                     'VISa',
                     'VISam']

def allen_get_raw_annotation(annotation_dir,
                             version = 'annotation/ccf_2017',
                             resolution = 10):
    import nrrd
    from allensdk.api.queries.mouse_connectivity_api import MouseConnectivityApi
    annotation_path = pjoin(annotation_dir, 'annotation_{0}_{1}.nrrd'.format(
        version,resolution))
    if not os.path.isdir(annotation_dir):
        os.makedirs(annotation_dir)
    if not os.path.isfile(annotation_path):
        mcapi = MouseConnectivityApi()
        mcapi.download_annotation_volume(version, resolution, annotation_path)
    annotation, meta = nrrd.read(annotation_path)
    return annotation,meta

def allen_volume_from_structures(
        structures = selection_dorsal_cortex,
        resolution=10,
        version = 'annotation/ccf_2017'):
    '''
    
    Gets specific regions from an annotation volume from the allen atlas

    '''
    from allensdk.api.queries.mouse_connectivity_api import MouseConnectivityApi
    from allensdk.api.queries.ontologies_api import OntologiesApi
    from allensdk.core.structure_tree import StructureTree
    from allensdk.core.reference_space import ReferenceSpace
    # the annotation download writes a file, so we will need somwhere to put it

    # the annotation download writes a file, so we will need somwhere to put it
    annotation, meta = allen_get_raw_annotation(
        annotation_dir,
        version = version,
        resolution = 10)
    oapi = OntologiesApi()
    structure_graph = oapi.get_structures_with_sets([1])  # 1 is the id of the adult mouse structure graph

    # This removes some unused fields returned by the query
    structure_graph = StructureTree.clean_structures(structure_graph)  
    tree = StructureTree(structure_graph)    
    rsp = ReferenceSpace(tree, annotation, [resolution]*3)

    areas = rsp.structure_tree.get_structures_by_acronym(structures)
    ids = [st['id'] for st in areas]
    mask_volume = np.zeros_like(annotation,dtype='int16')
    for i,sid in tqdm(enumerate(ids)):
        masks = rsp.make_structure_mask([sid])
        mask_volume[masks == 1] = i+1
        areas[i]['mask_volume_id'] = i+1
    return  mask_volume, areas
    
def allen_flatten_areas(areas,mask_volume,
                        resolution=10, reference= [540,570],gaussfilt=0):    
    ''' 
Creates a top view and extracts contours from each area in an annotation volume 
    proj, ccf_regions = allen_flatten_areas(areas, mask_volume, 
                                            resolution=10, 
                                            reference= [540 570])
    
    The default reference point is a (certainly wrong) guess of the location of bregma.

'''
    from .allen_utils import allen_top_proj_from_volume
    proj = allen_top_proj_from_volume(mask_volume)
    uproj = np.unique(proj)
    h,w = proj.shape
    proja = proj.copy()
    proja[:,-int(w/2):] = 0
    mask_ids = [a['mask_volume_id'] for a in areas]
    ccf_regions = []
    res = resolution/1000.
    for i,v in enumerate(mask_ids):
        # find the largest region for each brain area
        a = (proja==v)
        if not gaussfilt==0:
            a = gaussian(a,sigma=gaussfilt)
        a = measure.label(a)
        rprops = measure.regionprops(a)
        if not len(rprops):
            print('No projection for '+areas[i]['acronym'])
            continue
        ar = np.argmax([r.area for r in rprops])
        c = measure.find_contours(a==rprops[ar].label,.5)[0]
        cm = np.array(rprops[ar].centroid)
        left_center = cm-np.array(reference)
        d = c.copy()
        d[:,1] = -1*(c[:,1])+w
        right_center = [1,-1]*cm+w-np.array(reference)

        ccf_regions.append(dict(acronym=areas[i]['acronym'],
                                name=areas[i]['name'],
                                reference = reference,
                                resolution = resolution,
                                label = v,
                                allen_id = areas[i]['id'],
                                allen_rgb = areas[i]['rgb_triplet'],
                                left_center = left_center[::-1]*res,
                                right_center = right_center[::-1]*res,
                                left_x = (c[:,1] - reference[1])*res,
                                left_y = (c[:,0] - reference[0])*res,
                                right_x = (d[:,1] - reference[1])*res,
                                right_y = (d[:,0] - reference[0])*res))
    from pandas import DataFrame
    ccf_regions = DataFrame(ccf_regions)
    return proj,ccf_regions

def projection_outline(proj, resolution = 10, reference=[540,570]):
    '''
Get the outline from the projection

    outline_contour = projection_outline(proj, resolution = 0.010)
    
    '''
    
    a = ndimage.binary_closing(proj >0,np.ones((10,10)))
    a = ndimage.binary_fill_holes(a)
    a = ndimage.binary_dilation(a)
    return (measure.find_contours(a,.5)[0][:,::-1]- np.array(reference)[::-1])*resolution/1000. 



########################################################################
########################################################################
########################################################################

def allen_proj_extent(proj, ccf_regions, foraxis=False):
    '''
Utility function to return the extent of normalized axis for when plotting 
allen projections on matplotlib axes.

    extent = allen_proj_extent(proj, ccf_regions, foraxis=False)
    
    proj is the projection (used to get dims)
    ccf_refions is a dataframe of regions (to get the reference and resolution)
    foraxis = True flips the Y axis
    
Example:
    plt.imshow(proj,extent = allen_proj_extent(proj,ccf_regions),
               origin='top', cmap = 'gray')
    plt.plot(brainoutline[:,0],brainoutline[:,1],'c',lw=3)    
    plt.axis(allen_proj_extent(proj,ccf_regions, foraxis=True))

    '''
    h,w = proj.shape
    r2,r1 = ccf_regions.reference.iloc[0]
    extent = np.array([-r1,w-r1,-r2,h-r2])*ccf_regions.resolution.iloc[0]/1000.
    if foraxis:
        return extent[[0,1,3,2]]
    else:
        return extent

########################################################################
########################################################################
########################################################################


def apply_affine_to_points(x,y,M):
    '''
    Apply an affine transform to a set of contours or (x,y) points.

    x,y = apply_affine_to_points(x, y, tranform)

    '''
    if M is None:
        nM = np.identity(3,dtype = np.float32)
    else:
        nM = M.params
    xy = np.vstack([x,y,np.ones_like(y)])
    res = (nM @ xy).T
    return res[:,0],res[:,1]

def allen_transform_regions(M,ccf_regions,resolution = 1,bregma_offset = [0.,0.]):
    ''' This transforms regions from the reference to image coordinates.
    Usage:

    lmarks = load_allen_landmarks('dorsal_cortex_landmarks.json')
    refregions = allen_load_reference('dorsal_cortex')
    refregions_image = allen_transform_regions(lmarks['transform'],refregions,
                                        resolution = lmarks['resolution'],
                                        bregma_offset = lmarks['bregma_offset'])

    The output: refregions_image is the same as refregions but in image space.
 '''
    nccf = ccf_regions.copy()
    for i,c in nccf.iterrows():
        for side in ['left','right']:
            x,y = apply_affine_to_points(np.array(c[side+'_x'])/resolution + bregma_offset[0],
                                         np.array(c[side+'_y'])/resolution + bregma_offset[1], M)
            nccf.at[i,side+'_x'] = x.tolist()
            nccf.at[i,side+'_y'] = y.tolist()
            x,y = apply_affine_to_points(c[side+'_center'][0]/resolution + bregma_offset[0],
                                         c[side+'_center'][1]/resolution + bregma_offset[1], M)
            nccf.at[i,side+'_center'] = [x[0],y[0]]
    return nccf

def allen_transform_from_landmarks(landmarks_im,match):
    '''
    Compute the similarity transform from annotated landmarks. 
    
    transform = allen_transform_from_landmarks(landmarks_im,match)
    
    '''
    ref = np.vstack([landmarks_im['x'],landmarks_im['y']]).T
    cor = np.vstack([match['x'],match['y']]).T
    return estimate_similarity_transform(ref, cor)

def allen_landmarks_to_image_space(landmarks,
                                   bregma_offset = np.array([0,0]),
                                   resolution = 0.01):
    '''
    Convert landmarks from allen to "image" space.
    Basically just divides by the resolution and adds the bregma offset.

    Warning this does the operation in place, pass .copy()

    landmarks = allen_landmarks_to_image_space(landmarks.copy(),
                                   bregma_offset = np.array([0,0]),
                                   resolution = 0.01)
    '''
    landmarks['x'] = landmarks['x']/resolution + bregma_offset[0]
    landmarks['y'] = landmarks['y']/resolution + bregma_offset[1]
    return landmarks


#######################################################################
################      READ AND WRITE FUNCTIONS      ###################
#######################################################################

def save_allen_landmarks(landmarks,
                         filename = None,
                         resolution = None,
                         landmarks_match = None,
                         bregma_offset = None,
                         transform = None,
                         **kwargs):
    '''
    landmarks need to be pandas dataframes.

    default is dorsal_cortex_landmarks.json in the ~/.wfield directory.

    '''
    lmarks = dict(landmarks=landmarks.to_dict(orient='list'))
    if not resolution is None:
        lmarks['resolution'] = resolution    
    if not landmarks_match is None:
        lmarks['landmarks_match'] = landmarks_match.to_dict(orient='list')
    if not bregma_offset is None:
        if isinstance(bregma_offset,np.ndarray):
            bregma_offset = bregma_offset.tolist()
        lmarks['bregma_offset'] = bregma_offset
    if not transform is None:
        from skimage.transform import SimilarityTransform
        if isinstance(transform,SimilarityTransform):
            lmarks['transform'] = transform.params.tolist()
        elif isinstance(transform,np.ndarray):
            lmarks['transform'] = transform.tolist()
        else:
            lmarks['transform'] = transform
    if 'bregma_offset' in lmarks.keys() and 'resolution' in lmarks.keys():
        lmarks['landmarks_im'] = allen_landmarks_to_image_space(
            landmarks.copy(), 
            lmarks['bregma_offset'],
            lmarks['resolution']).to_dict(orient='list')
    if filename is None:
        filename = pjoin(annotation_dir,'dorsal_cortex_landmarks.json')
    with open(filename,'w') as fd:
        import json
        json.dump(lmarks,fd, sort_keys = True, indent = 4)

def load_allen_landmarks(filename, reference = 'dorsal_cortex'):
    if filename is None:
        filename = pjoin(annotation_dir,reference + '_landmarks.json')
    if not os.path.exists(filename):
        if '.wfield' in filename:
            from .utils import _create_wfield_folder
            _create_wfield_folder()
        else:
            raise(OSError('Could not find the reference file {0}.'.format(filename)))

    with open(filename,'r') as fd:
        import json
        lmarks = json.load(fd)
    for k in ['landmarks_im','landmarks','landmarks_match']:
        if k in lmarks.keys():
            from pandas import DataFrame
            lmarks[k] = DataFrame(lmarks[k])[['x','y','name','color']]
    if 'transform' in lmarks.keys():
        from skimage.transform import SimilarityTransform
        lmarks['transform'] = SimilarityTransform(
            np.array(lmarks['transform']))
    return lmarks

    
def allen_load_reference(reference_name, annotation_dir = annotation_dir):
    '''
Load allen areas to use as reference.

Example:
    ccf_regions,proj,brain_outline = allen_load_reference('dorsal_cortex')

    '''
    if annotation_dir == wfield_dir:
        # then it is the reference folder, download if not there
        if not os.path.exists(pjoin(
                annotation_dir,
                '{0}_ccf_labels.json'.format(reference_name))):
            from .utils import _create_wfield_folder
            _create_wfield_folder()
    from pandas import read_json
    ccf_regions = read_json(pjoin(
        annotation_dir,'{0}_ccf_labels.json'.format(reference_name)))
    proj = np.load(pjoin(annotation_dir,
                         '{0}_projection.npy'.format(reference_name),))
    brain_outline = np.load(pjoin(annotation_dir,
                                  '{0}_outline.npy'.format(reference_name)))
    return ccf_regions,proj,brain_outline

def allen_save_reference(ccf_regions,
                         proj,
                         brainoutline,
                         reference_name,
                         annotation_dir=annotation_dir):
    '''
    Save Allen references to the default directory

    allen_save_reference(ccf_regions, proj, brainoutline,
                         referece_name,
                         annotation_dir=annotation_dir):
    '''
    
    ccf_regions.to_json(pjoin(
        annotation_dir,'{0}_ccf_labels.json'.format(reference_name)),
                        orient='records')
    np.save(pjoin(annotation_dir,
                  '{0}_projection.npy'.format(reference_name)),proj)
    np.save(pjoin(annotation_dir,
                  '{0}_outline.npy'.format(reference_name)),brainoutline)

    

def allen_regions_to_atlas(ccf_regions,dims,
                           sides = ['left','right'],
                           fillnan = False):
    ''' 
    Atlas as called in locaNMF; it is the masks of allen areas.
    This function returns also the names of the areas
    '''
    atlas = np.zeros(dims,dtype = np.float32)
    if fillnan:
        atlas.fill(np.nan)
    areanames = []
    for ireg,r in ccf_regions.iterrows():
        for iside,side in enumerate(sides):
            mask = contour_to_mask(
                r[side+'_x'],r[side+'_y'],
                dims = dims)
            factor = 1
            if iside==1:
                factor = -1
            atlas[mask==1] = factor*(ireg+1)
            areanames.append([factor*(ireg+1),r['acronym']+'_'+side])
    return atlas,areanames

def atlas_from_landmarks_file(landmarks_file=None, reference='dorsal_cortex', dims = [540,640], do_transform = None):
    lmarks = load_allen_landmarks(landmarks_file)
    ccf_regions,proj,brain_outline = allen_load_reference('dorsal_cortex')
    # transform the regions into the image
    if not 'transform' in lmarks.keys():
        lmarks['transform'] = None
    transform = lmarks['transform']
    if not do_transform:
        transform = None
    nccf_regions = allen_transform_regions(transform,
                                           ccf_regions,
                                           resolution=lmarks['resolution'],
                                           bregma_offset=lmarks['bregma_offset'])
    nbrain_outline = apply_affine_to_points(brain_outline[:,0]/lmarks['resolution'] + lmarks['bregma_offset'][0],
                                            brain_outline[:,1]/lmarks['resolution'] + lmarks['bregma_offset'][1],
                                            transform)


    atlas,areanames = allen_regions_to_atlas(nccf_regions, dims)
    brain_mask = contour_to_mask(*nbrain_outline, dims = dims)
    return atlas, areanames, brain_mask

########################################################################
################HOLOVIEWS PLOTTING FUNCTIONS############################
########################################################################

def hv_plot_allen_regions(ccf_regions,
                          resolution = 1,
                          bregma_offset = np.array([0,0]),
                          side_selection='both'):
    '''
    Example: 

import holoviews as hv
hv.extension('bokeh')

ccf_regions,proj,brain_outline = allen_load_reference('dorsal_cortex')
hv_plot_allen_regions(ccf_regions).options({'Curve': {'color':'black', 'width': 600}})
    
    '''
    import holoviews as hv
    regs = []
    for p in ccf_regions.iterrows():
        c = p[1]
        if side_selection in ['right','both']:
            regs.append(hv.Curve(np.vstack([np.array(c.right_x)/resolution + bregma_offset[0],
                                            np.array(c.right_y)/resolution + bregma_offset[1]]).T))
        if side_selection in ['left','both']:
            regs.append(hv.Curve(np.vstack([np.array(c.left_x)/resolution + bregma_offset[0],
                                            np.array(c.left_y)/resolution + bregma_offset[1]]).T))
    # return a plot
    plot = regs[0]
    for i in range(1,len(regs)):
        plot *= regs[i]
    return plot.opts(invert_yaxis=True).options(width = 600)

def hv_adjust_reference_landmarks(landmarks,ccf_regions,msize=40):
    '''
    landmarks = {'x': [-1.95, 0, 1.95, 0],
                 'y': [-3.45, -3.45, -3.45, 3.2],
                 'name': ['OB_left', 'OB_center', 'OB_right', 'RSP_base'],
                 'color': ['#fc9d03', '#0367fc', '#fc9d03', '#fc4103']}
    landmarks = pd.DataFrame(landmarks)
    # adjust landmarks
    wid,landmark_wid = hv_adjust_reference_landmarks(landmarks,ccf_regions)
    wid # to display
    # use the following to retrieve (on another cell) 
    landmarks = pd.DataFrame(landmark_wid.data)[['x','y','name','color']]
    '''
    import holoviews as hv
    from holoviews import opts, streams
    from holoviews.plotting.links import DataLink
    referenceplt = hv_plot_allen_regions(ccf_regions).options(
        {'Curve': {'color':'black', 'width': 600}})

    points = hv.Points(landmarks,vdims='color').opts(marker='+',size=msize)
    point_stream = streams.PointDraw(data=points.columns(), 
                                     add = False,num_objects=4, 
                                     source=points, empty_value='black')
    table = hv.Table(points, ['x', 'y','name'], 'color').opts(title='Landmarks location')
    DataLink(points, table)
    widget = (referenceplt*points + table).opts(
        opts.Layout(merge_tools=False),
        opts.Points(invert_yaxis=True,
                    active_tools=['point_draw'],
                    color='color', height=500,
                    tools=['hover'], width=500),
        opts.Table(editable=True))
    return widget,point_stream

def hv_adjust_image_landmarks(image,landmarks,
                              landmarks_match = None,
                              bregma_offset = None,
                              resolution = 0.0194,
                              msize = 40):
    '''
    TODO: merge part of this with the one for the landmarks
    landmarks are in allen reference space
    '''
    h,w = image.shape
    if bregma_offset is None:
        # then it is the center of the image
        bregma_offset = np.array([int(w/2),int(h/2)]) # place bregma in the center of the image
        
    landmarks_im = allen_landmarks_to_image_space(landmarks.copy(),bregma_offset,resolution)
    if landmarks_match is None:
        landmarks_match = landmarks_im
    import holoviews as hv
    from holoviews import opts, streams
    from holoviews.plotting.links import DataLink

    bounds = np.array([0,0,w,h])
    im = hv.Image(image[::-1,:],
                 bounds =tuple(bounds.tolist())).opts(
        invert_yaxis = True,cmap = 'gray')

    points = hv.Points(landmarks_match,vdims='color').opts(marker='+',size=msize)
    point_stream = streams.PointDraw(data=points.columns(), 
                                     add = False,num_objects=4, 
                                     source=points, empty_value='black')
    table = hv.Table(points, ['x', 'y','name'], 'color').opts(title='Annotation location')
    DataLink(points, table)

    from bokeh.models import HoverTool
    hoverpts = HoverTool(tooltips=[("i", "$index")])

    widget = (im*points + table).opts(
        opts.Layout(merge_tools=False),
        opts.Points(invert_yaxis=True,active_tools=['point_draw'], 
                    color='color',
                    tools=[hoverpts], 
                    width=int(w),
                    height=int(h)),
        opts.Table(editable=True))
    return widget,point_stream,landmarks_im

def hv_show_transformed_overlay(image, M, ccf_regions,
                                bregma_offset = None,resolution=0.0194):
    import holoviews as hv
    h,w = image.shape
    if bregma_offset is None:
        bregma_offset = np.array([int(w/2),int(h/2)]) # place bregma in the center of the image
    bounds = np.array([0,0,w,h])
    warped = warp(image,M,
                  order = 0,
                  clip=True,
                  preserve_range=True)
    im=hv.Image(warped[::-1,:],
                 bounds =tuple(bounds.tolist())).opts(invert_yaxis = True,cmap = 'gray')
    return im*hv_plot_allen_regions(ccf_regions,
                                    bregma_offset=bregma_offset,
                                    resolution=resolution).options(
        {'Curve': {'color':'white', 'width': w,'height':h}})





#  wfield - tools to analyse widefield data - allen utils (numba) 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from numba import jit
from numba import int16 as numba_int16
import numpy as np

@jit(nopython=True)
def allen_top_proj_from_volume(bvol):
    '''
Get the top projection from a volume.

    proj = allen_top_proj_from_volume(allen_volume)

'''
    h,d,w = bvol.shape
    proj = np.zeros((h,w),dtype=numba_int16)
    # this can be done with a np.where but is probably faster like this
    for i in range(h):
        for j in range(w):
            for z in range(d): 
                if bvol[i,z,j] > 0:
                    proj[i,j] = bvol[i,z,j]
                    break
            
    return proj



#  wfield - tools to analyse widefield data - dask tools 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .io import *
from dask import array as da
from dask import delayed
from dask.array import stats as dastats

def dask_load_binary(binfile,
                     nchan=None,
                     w=None,
                     h=None,
                     dtype=None,
                     blocksize = 512):
    assert os.path.isfile(binfile)
    meta = os.path.splitext(binfile)[0].split('_')
    if nchan is None: # try to get it from the filename
        nchan = int(meta[-4])
    if w is None: # try to get it from the filename
        w = int(meta[-3])
    if h is None: # try to get it from the filename
        h = int(meta[-2])
    if dtype is None: # try to get it from the filename
        dtype = meta[-1]
    dt = np.dtype(dtype)
    nframes = os.path.getsize(binfile)/(nchan*w*h*dt.itemsize)
    iblocks = np.arange(0,nframes,blocksize,dtype=int)
    if iblocks[-1] < nframes:
        iblocks = np.append(iblocks,nframes)
    framesize = int(nchan*w*h)
    blocks = []
    for ii,offset in enumerate(iblocks[:-1]):
        bsize = int(iblocks[ii+1] - offset)
        blocks.append((binfile,offset,bsize))
    lazyload = delayed(lambda x: load_binary_block(x,shape=(nchan,w,h)), 
                       pure=True)  # Lazy version of sbb_load_block
    sample = load_binary_block(blocks[0],shape=(nchan,w,h))
    lazy_values = [lazyload(block) for block in blocks]
    arrays = [da.from_delayed(lazy_value,           
                              dtype=sample.dtype,   
                              shape=[block[-1]] + [s for s in sample.shape[1:]])
          for block,lazy_value in zip(blocks,lazy_values)]
    return da.concatenate(arrays, axis=0)



#  wfield - tools to analyse widefield data - command line interface 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *
from .io import load_stack, mmap_dat
from .io import frames_average_for_trials
from .registration import motion_correct
from .decomposition import approximate_svd
from .hemocorrection import hemodynamic_correction
import argparse
import subprocess
import shlex
import platform
from multiprocessing import set_start_method

class CLIParser(object):
    def __init__(self):
        parser = argparse.ArgumentParser(
            description='wfield - tools to process widefield data',
            usage='''wfield <command> [<args>]

The commands are:
    ncaas               Opens a gui to launch data on the neuroCAAS platform

    open                Opens a gui to look at the preprocessed data        
    open_raw            Opens a gui to look at the raw frames

    preprocess          Preprocess stack 

    motion              Registers stack
    decompose           Performs single value decomposition
    correct             Performs hemodynamic correction on dual channel data

    imager              
    imager_preprocess   Preprocesses data recorded with the WidefieldImager
''')
        parser.add_argument('command', help='type wfieldtools <command> -h for help')

        args = parser.parse_args(sys.argv[1:2])
        set_start_method("spawn")
        if not hasattr(self, args.command):
            print('Command {0} not recognized.'.format(args.command))
            parser.print_help()
            exit(1)
        getattr(self, args.command)()

    def ncaas(self):
        parser = argparse.ArgumentParser(
            description='Open the GUI to interact with neuroCAAS.org',
            usage = '''
You'll need credentials from neurocaas.org before you are able to use this.

Type wfield ncaas <foldername> to open on a specific folder.
            ''')
        parser.add_argument('--foldername', action='store',default='.',type=str)
        args = parser.parse_args(sys.argv[2:])
        
        from wfield.ncaas_gui import main as ncaas_gui

        ncaas_gui(args.foldername)
        
        
    def open(self):
        parser = argparse.ArgumentParser(
            description='Open the GUI to look at preprocessed data',
            usage = '''

            Press control while moving the mouse around to look at the timecourse at different points.

            Right click and drag mouse on the trace plot to zoom.
''')
        parser.add_argument('foldername', action='store',default=None,type=str)
        parser.add_argument('--allen-reference', action='store',default='dorsal_cortex',type=str)
        parser.add_argument('--no-ncaas', action='store_true',default=False)
        parser.add_argument('--correlation', action='store_true',default=False,help = 'Show the correlation window - dont do this with sparse matrices.')
        parser.add_argument('--nchannels', action='store', default=None, type=int)
        parser.add_argument('--before-corr', action='store_true',
                            default=False,
                            help= 'Load SVT before hemodynamics correction.')
        args = parser.parse_args(sys.argv[2:])
        localdisk = args.foldername
        from .io import mmap_dat
        fname = pjoin(localdisk,'U.npy')
        Uwarped = None
        if os.path.isdir(pjoin(localdisk,'results')) and not args.no_ncaas:
            # then it is from neurocaas?
            fname = pjoin(localdisk,'results','U_atlas.npy')
            if os.path.isfile(fname):
                Uwarped = np.load(fname)
                print('Found U_atlas')
                iswarped = True
            fname = pjoin(localdisk,'results','config.yaml')
            if os.path.isfile(fname):
                with open(fname,'r') as f:
                    import yaml
                    config = yaml.safe_load(f)
                H,W = (config['fov_height'],config['fov_width'])
                fname = pjoin(localdisk,'results','reduced_spatial.npy')
                if os.path.isfile(fname):
                    U = np.load(fname)
                    U = U.reshape([H,W,-1])
                    dims = [H,W]
            fname = pjoin(localdisk,'results','SVTcorr.npy')
            if not os.path.isfile(fname):
                fname = pjoin(localdisk,'results','reduced_temporal.npy')
            SVT = np.load(fname)
        elif os.path.isfile(fname):
            U = np.load(fname)
            dims = U.shape[:2]
            if (not args.before_corr
                and os.path.isfile(pjoin(localdisk,'SVTcorr.npy'))):
                SVT = np.load(pjoin(localdisk,'SVTcorr.npy'))
            else:
                SVT = np.load(pjoin(localdisk,'SVT.npy'))
        else:
            # try in a results folder (did it come from ncaas?)
            raise OSError('Could not find: {0} '.format(fname))

        dat = load_stack(args.foldername, nchannels = args.nchannels)
        #dat_path = glob(pjoin(localdisk,'*.bin'))
        #if len(dat_path):
        #    dat_path = dat_path[0]
        #    if os.path.isfile(dat_path):
        #        dat = mmap_dat(dat_path)
        if dat is None:
            dat = None
            dat_path = None
            average_path = pjoin(localdisk,'frames_average.npy')
            if os.path.isfile(average_path):
                dat = np.load(average_path)
                dat = dat.reshape([1,*dat.shape])

        trial_onsets = pjoin(localdisk,'trial_onsets.npy')
        if os.path.isfile(trial_onsets):
            trial_onsets = np.load(trial_onsets).astype(int)
        else:
            trial_onsets = None
        
        stack = SVDStack(U,SVT, warped = Uwarped, dims = dims)
        if not Uwarped is None:
            stack.set_warped(True)
        from .widgets import QApplication,SVDViewer
        app = QApplication(sys.argv)
        w = SVDViewer(stack,
                      folder = localdisk,
                      raw = dat,
                      trial_onsets = trial_onsets,
                      reference = args.allen_reference,
                      start_correlation = args.correlation)
        sys.exit(app.exec_())
        del dat
        
    def open_raw(self):
        parser = argparse.ArgumentParser(
            description='Inspect the raw video frames')
        parser.add_argument('foldername', action='store',
                            default=None, type=str,
                            help='Folder where to search for files.')
        parser.add_argument('--nchannels', action='store', default=None, type=int)
        parser.add_argument('--allen-reference', action='store',default='dorsal_cortex',type=str)
        parser.add_argument('--napari', action='store_true',
                            default=False,
                            help='Show with napari')
        args = parser.parse_args(sys.argv[2:])
        localdisk = args.foldername
        dat = load_stack(args.foldername, nchannels = args.nchannels)

        if os.path.isdir(localdisk):
            dat_path = glob(pjoin(localdisk,'*.bin'))
            if len(dat_path):
                dat_path = dat_path[0]
        else:
            dat_path = localdisk
            localdisk = os.path.dirname(localdisk)

        if args.napari:
            from .viz import napari_show
            napari_show(dat)
            del dat
            sys.exit()

        if dat is None:
            average_path = pjoin(localdisk,'frames_average.npy')
            if os.path.isfile(average_path):
                dat = np.load(average_path)
            dat = dat.reshape([1,*dat.shape])
            print('Loading the frames_average instead')
        trial_onsets = pjoin(localdisk,'trial_onsets.npy')
        if os.path.isfile(trial_onsets):
            trial_onsets = np.load(trial_onsets).astype(int)
        else:
            trial_onsets = None
        from .widgets import QApplication,RawViewer
        app = QApplication(sys.argv)
        w = RawViewer(raw = dat,
                      folder = localdisk,
                      trial_onsets = trial_onsets,
                      reference = args.allen_reference)
        sys.exit(app.exec_())
        del dat

        
    def imager(self):
        parser = argparse.ArgumentParser(
            description='Converts widefield data recorded with the WidefieldImager')
        parser.add_argument('foldername', action='store',
                            default=None, type=str,
                            help='Folder where to search for mj2 and analog files.')
        parser.add_argument('-o','--output', action='store',
                            default=None, type=str,
                            help='Output folder') # there should be an intermediate folder as well
        
        args = parser.parse_args(sys.argv[2:])
        remotepath = args.foldername
        localdisk = args.output # this should be an SSD or a fast drive

        if localdisk is None:
            print('Specify a fast local disk with the  -o option.')
            exit(1)
        if not os.path.isdir(localdisk):
            os.makedirs(localdisk)
            print('Created {0}'.format(localdisk))
        print('''

    Searching for mj2 files in: 
        {0}
        
    Output folder: 
        {1}

        '''.format(remotepath,localdisk))
        tfetch = time.time()
        dat,frames_avg, trialonsets,trialinfo = parse_imager_mj2_folder(remotepath, localdisk)
        del dat
        del frames_avg
        del trialonsets
        del trialinfo
        tfetch = (time.time() - tfetch)/60.

        
    def imager_preprocess(self):
        parser = argparse.ArgumentParser(
        description='Performs preprocessing of widefield data recorded with the WidefieldImager')
        parser.add_argument('foldername', action='store',
                            default=None, type=str,
                            help='Folder where to search for mj2 and analog files.')
        parser.add_argument('-o','--output', action='store',
                            default=None, type=str,
                            help='Output folder') # there should be an intermediate folder as well
        parser.add_argument('-k', action='store',default=200,type=int,
                            help = 'Number of components for SVD')
        #parser.add_argument('--mask-edge', action='store',default=30,type=int,
        #                    help = 'Size of the mask used on the edges during motion correction ') 
        parser.add_argument('--nbaseline-frames', action='store',
                            default=30, type=int,
                            help='Number of frames to compute the  baseline')
        parser.add_argument('--fs', action='store',default=30.,type=np.float32,
                            help='Sampling frequency of an individual channel')
        parser.add_argument('--mode', choices=('ecc','2d'),default='ecc',
                            help = 'Algorithm for  motion correction ')
        parser.add_argument('--chunksize', default=256,
                            help = 'Frames per batch ') 


        args = parser.parse_args(sys.argv[2:])
        remotepath = args.foldername
        localdisk = args.output # this should be an SSD or a fast drive

        if localdisk is None:
            print('Specify a fast local disk with the  -o option.')
            exit(1)
        if not os.path.isdir(localdisk):
            os.makedirs(localdisk)
            print('Created {0}'.format(localdisk))
        print('''

    Searching for mj2 files in: 
        {0}
        
    Output folder: 
        {1}

        '''.format(remotepath,localdisk))
        tfetch = time.time()
        dat,frames_avg, trialonsets,trialinfo = parse_imager_mj2_folder(remotepath, localdisk)
        del dat
        del frames_avg
        del trialonsets
        del trialinfo
        tfetch = (time.time() - tfetch)/60.

        tproc = time.time()
        # MOTION CORRECTION
        _motion(localdisk,
                outdisk = localdisk,
                chunksize = args.chunksize,
                mode = args.mode)
        # COMPUTE AVERAGE FOR BASELINE
        _baseline(localdisk,args.nbaseline_frames)
        # DATA REDUCTION
        _decompose(localdisk, k=args.k)
        # HEMODYNAMIC CORRECTION
        _hemocorrect(localdisk,fs=args.fs)
        tproc = (time.time() - tproc)/60.
        print('Done fetching data ({0} min) and pre-processing ({1} min)'.format(tfetch,tproc))
        exit(0)

    def preprocess(self):
        parser = argparse.ArgumentParser(
            description='Performs preprocessing of widefield data recorded from a binary file',
            usage = '''
            wfield preprocess FOLDERNAME

     The folder must contain a binary file; the end of the filename must be _NCHANNELS_H_W_DTYPE.dat
     
           Example name: frames_2_540_640_uint16.dat
    '''
)
        parser.add_argument('foldername', action='store',
                            default=None, type=str,
                            help='Folder that has the binary file (FAST DISK).')
        parser.add_argument('-o','--output', action='store',
                            default='wfield_results', type=str,
                            help='Output folder')
        parser.add_argument('-k', action='store',default=200,type=int,
                            help = 'Number of components for SVD')
        #parser.add_argument('--mask-edge', action='store',default=30,type=int,
        #                    help = 'Size of the mask used on the edges during motion correction ') 
        parser.add_argument('--nchannels', action='store', default=None, type=int,
                            help = 'Number of recorded channels (interleaved)')
        parser.add_argument('--functional-channel', action='store',
                            default=0, type=int,
                            help='Index of the functional channel')
        parser.add_argument('--nbaseline-frames', action='store',
                            default=30, type=int,
                            help='Number of frames to compute the  baseline')
        parser.add_argument('--fs', action='store',default=30.,type=np.float32,
                            help='Sampling frequency of an individual channel')
        parser.add_argument('--mode', choices=('ecc','2d'),default='ecc',
                            help = 'Algorithm for  motion correction ') 
        parser.add_argument('--chunksize', default=256,
                            help = 'Frames per batch ') 

        
        args = parser.parse_args(sys.argv[2:])
        datadisk = os.path.abspath(args.foldername) # this should be an SSD or a fast drive
        
        localdisk = args.output # this should be an SSD or a fast drive
        if localdisk is None:
            print('Specify a fast local disk with the  -o option.')
            exit(1)
        if localdisk == 'wfield_results':
            localdisk = pjoin(datadisk,localdisk)
        else:
            localdisk = os.path.abspath(localdisk)
        if not os.path.isdir(localdisk):
            os.makedirs(localdisk)
            print('Created {0}'.format(localdisk))

        tproc = time.time()
        # MOTION CORRECTION
        _motion(datadisk,outdisk = localdisk,
                mode = args.mode,
                chunksize = args.chunksize,
                nchannels = args.nchannels)
        # COMPUTE AVERAGE FOR BASELINE
        _baseline(localdisk,args.nbaseline_frames, nchannels = args.nchannels)
        # DATA REDUCTION
        _decompose(localdisk, k=args.k, nchannels = args.nchannels)
        # HEMODYNAMIC CORRECTION
        # check if it is 2 channel
        dat = load_stack(localdisk, nchannels = args.nchannels)
        if dat.shape[1] == 2:
            del dat
            _hemocorrect(localdisk,fs=args.fs, functional_channel = args.functional_channel)
            
        tproc = (time.time() - tproc)/60.
        print('Done  pre-processing ({0} min)'.format(tproc))
        exit(0)
        
    def motion(self):
        parser = argparse.ArgumentParser(
            description='Performs motion correction on widefield data')
        parser.add_argument('foldername', action='store',
                            default=None, type=str,
                            help='Folder with dat file.')
        parser.add_argument('--nchannels', action='store', default=None, type=int)
        parser.add_argument('--mode', choices=('ecc','2d'),default='ecc',
                            help = 'Algorithm for  motion correction ') 
        parser.add_argument('--chunksize', default=256,
                            help = 'Frames per batch ') 
        parser.add_argument('--mode', choices=('ecc','2d'),default='ecc',
                            help = 'Algorithm for  motion correction ') 

        args = parser.parse_args(sys.argv[2:])
        localdisk = args.foldername
        _motion(localdisk,
                nchannels = args.nchannels,
                mode = args.mode,
                chunksize = args.chunksize)

    def decompose(self):
        parser = argparse.ArgumentParser(
            description='Performs single value decomposition')
        parser.add_argument('foldername', action='store',
                            default=None, type=str,
                            help='Folder with dat file.')
        parser.add_argument('-k', action='store',default=200,type=int,
                            help = 'Number of components for SVD ') 
        parser.add_argument('--nchannels', action='store', default=None, type=int)
        parser.add_argument('--no-baseline',
                            action='store_true', default=False,
                            help = 'Skip baseline ') 
        parser.add_argument('--nbaseline-frames', action='store',
                            default=30, type=int,
                            help='Number of frames to compute the  baseline')        
        args = parser.parse_args(sys.argv[2:])
        localdisk = args.foldername
        if not args.no_baseline:
            _baseline(localdisk,args.nbaseline_frames, nchannels = args.nchannels)
        _decompose(localdisk,k=args.k, nchannels = args.nchannels)

    def correct(self):
        parser = argparse.ArgumentParser(
            description='Performs hemodynamics correction (needs 2 channel compressed data)')
        parser.add_argument('foldername', action='store',
                            default=None, type=str,
                            help='Folder with U and SVT files.')
        parser.add_argument('--fs', action='store',default=30.,type=np.float32,
                            help = 'Sampling rate of an individual channel ') 
        parser.add_argument('--functional-channel', action='store',
                            default=0, type=int,
                            help='Index of the functional channel')
        
        args = parser.parse_args(sys.argv[2:])
        localdisk = args.foldername 

        _hemocorrect(localdisk, fs=args.fs,
                     functional_channel=args.functional_channel)
        
def _motion(localdisk,
            nchannels = None,
            mode = 'ecc',
            chunksize=256,
            outdisk = None,
            in_place = False):
    # TODO: check if a motion corrected file is already here
    # if a single binary file try in-place
    dat_path = glob(pjoin(localdisk,'*.bin'))
    if outdisk is None:
        outdisk = localdisk
    if not len(dat_path):
        dat_path = glob(pjoin(localdisk,'*.dat'))
    if len(dat_path) == 1:
        dat = mmap_dat(dat_path[0], mode='r+')
    else:
        # else do else
        dat = load_stack(localdisk, nchannels = nchannels)
    if in_place:
        out = dat
    else:
        out = np.memmap(pjoin(outdisk,'motioncorrect_{0}_{1}_{2}_{3}.bin'.format(*dat.shape[1:],dat.dtype)),
                        mode='w+',
                        dtype=dat.dtype,
                        shape = dat.shape)
    (yshifts,xshifts),rshifts = motion_correct(dat,
                                               out=out,
                                               chunksize=int(chunksize),
                                               mode = mode,
                                               apply_shifts=True)
    del out # close and finish writing
    shifts = np.rec.array([yshifts,xshifts],dtype=[('y','float32'),('x','float32')])
    np.save(pjoin(outdisk,'motion_correction_shifts.npy'),shifts)
    np.save(pjoin(outdisk,'motion_correction_rotation.npy'),rshifts)
    from .plots import plot_summary_motion_correction
    plot_summary_motion_correction(shifts,outdisk)
    del shifts

def _baseline(localdisk, nbaseline_frames, nchannels = None):

    if os.path.exists(pjoin(localdisk,'frames_average.npy')):
        print('Found frame_average.npy. skipping.')
        
        return np.load(pjoin(localdisk,'frames_average.npy'))
    dat = load_stack(localdisk,nchannels = nchannels)
    try:
        trial_onsets = np.load(pjoin(localdisk,'trial_onsets.npy'))[:,1].astype(int)
    except FileNotFoundError:
        print('''
 Estimating the mean by the average of the chunked mean projection.
''')
        chunkidx = chunk_indices(len(dat),chunksize=64)
        frame_averages = []
        for on,off in tqdm(chunkidx):
            frame_averages.append(dat[on:off].mean(axis=0))
        del dat
        frames_average = np.stack(frame_averages).mean(axis = 0)
        np.save(pjoin(localdisk,'frames_average.npy'),
                frames_average)
        return frames_average
    frames_average_trials = frames_average_for_trials(dat,
                                                      trial_onsets,
                                                      nbaseline_frames)
    
    np.save(pjoin(localdisk,'frames_average.npy'),
            frames_average_trials.mean(axis=0))
    del dat
    del frames_average_trials

def _decompose(localdisk, k, nchannels = None):
    dat = load_stack(localdisk,nchannels = nchannels)

    frames_average = np.load(pjoin(localdisk,'frames_average.npy'))
    if len(frames_average)>3:
        trial_onsets = np.load(pjoin(localdisk,'trial_onsets.npy'))
        onsets = trial_onsets[:,1].astype(int)

    else:
        onsets = None
    U,SVT = approximate_svd(dat, frames_average, onsets = onsets, k = k)
    np.save(pjoin(localdisk,'U.npy'),U)
    np.save(pjoin(localdisk,'SVT.npy'),SVT)

def _hemocorrect(localdisk,fs,functional_channel = 0):
    U = np.load(pjoin(localdisk,'U.npy'))
    SVT = np.load(pjoin(localdisk,'SVT.npy'))
    
    SVT_470 = SVT[:,np.mod(functional_channel,2)::2]
    t = np.arange(SVT.shape[1]) # interpolate the violet
    from scipy.interpolate import interp1d
    SVT_405 = interp1d(t[np.mod(functional_channel+1,2)::2],
                       SVT[:,np.mod(functional_channel+1,2)::2],axis=1,
                       fill_value='extrapolate')(t[0::2])
    SVTcorr, rcoeffs, T = hemodynamic_correction(U, SVT_470, SVT_405, fs=fs)  

    np.save(pjoin(localdisk,'rcoeffs.npy'),rcoeffs)
    np.save(pjoin(localdisk,'T.npy'),T)
    np.save(pjoin(localdisk,'SVTcorr.npy'),SVTcorr)
    from .plots import plot_summary_hemodynamics_dual_colors
    plot_summary_hemodynamics_dual_colors(rcoeffs,
                                          SVT_470,
                                          SVT_405,
                                          U,
                                          T,
                                          frame_rate=fs,
                                          duration = 12,
                                          outputdir = localdisk)            
    
def main():
    CLIParser()

if __name__ == '__main__':
    main()



#  wfield - tools to analyse widefield data - IO tools 
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *

def load_dat(filename,
             nframes = None,
             offset = 0,
             shape = None,
             dtype='uint16'): 
    '''
    Loads frames from a binary file.
    
    Inputs:
        filename (str)       : fileformat convention, file ends in _NCHANNELS_H_W_DTYPE.dat
        nframes (int)        : number of frames to read (default is None: the entire file)
        offset (int)         : offset frame number (default 0)
        shape (list|tuple)   : dimensions (NCHANNELS, HEIGHT, WIDTH) default is None
        dtype (str)          : datatype (default uint16) 
    Returns:
        An array with size (NFRAMES,NCHANNELS, HEIGHT, WIDTH).

    Example:
        dat = load_dat(filename)
        
    ''' 
    if not os.path.isfile(filename):
        raise OSError('File {0} not found.'.format(filename))
    if shape is None or dtype is None: # try to get it from the filename
        dtype,shape,_ = _parse_binary_fname(filename,
                                            shape = shape,
                                            dtype = dtype)
    if type(dtype) is str:
        dt = np.dtype(dtype)
    else:
        dt = dtype

    if nframes is None:
        # Get the number of samples from the file size
        nframes = int(os.path.getsize(filename)/(np.prod(shape)*dt.itemsize))
    framesize = int(np.prod(shape))

    offset = int(offset)
    with open(filename,'rb') as fd:
        fd.seek(offset*framesize*int(dt.itemsize))
        buf = np.fromfile(fd,dtype = dt, count=framesize*nframes)
    buf = buf.reshape((-1,*shape),
                      order='C')
    return buf

def _parse_binary_fname(fname,lastidx=None, dtype = 'uint16', shape = None, sep = '_'):
    '''
    Gets the data type and the shape from the filename 
    This is a helper function to use in load_dat.
    
    out = _parse_binary_fname(fname)
    
    With out default to: 
        out = dict(dtype=dtype, shape = shape, fnum = None)
    '''
    fn = os.path.splitext(os.path.basename(fname))[0]
    fnsplit = fn.split(sep)
    fnum = None
    if lastidx is None:
        # find the datatype first (that is the first dtype string from last)
        lastidx = -1
        idx = np.where([not f.isnumeric() for f in fnsplit])[0]
        for i in idx[::-1]:
            try:
                dtype = np.dtype(fnsplit[i])
                lastidx = i
            except TypeError:
                pass
    if dtype is None:
        dtype = np.dtype(fnsplit[lastidx])
    # further split in those before and after lastidx
    before = [f for f in fnsplit[:lastidx] if f.isdigit()]
    after = [f for f in fnsplit[lastidx:] if f.isdigit()]
    if shape is None:
        # then the shape are the last 3
        shape = [int(t) for t in before[-3:]]
    if len(after)>0:
        fnum = [int(t) for t in after]
    return dtype,shape,fnum


def mmap_dat(filename,
             mode = 'r',
             nframes = None,
             shape = None,
             dtype='uint16'):
    '''
    Loads frames from a binary file as a memory map.
    This is useful when the data does not fit to memory.
    
    Inputs:
        filename (str)       : fileformat convention, file ends in _NCHANNELS_H_W_DTYPE.dat
        mode (str)           : memory map access mode (default 'r')
                'r'   | Open existing file for reading only.
                'r+'  | Open existing file for reading and writing.                 
        nframes (int)        : number of frames to read (default is None: the entire file)
        offset (int)         : offset frame number (default 0)
        shape (list|tuple)   : dimensions (NCHANNELS, HEIGHT, WIDTH) default is None
        dtype (str)          : datatype (default uint16) 
    Returns:
        A memory mapped  array with size (NFRAMES,NCHANNELS, HEIGHT, WIDTH).

    Example:
        dat = mmap_dat(filename)
    '''
    
    if not os.path.isfile(filename):
        raise OSError('File {0} not found.'.format(filename))
    if shape is None or dtype is None: # try to get it from the filename
        dtype,shape,_ = _parse_binary_fname(filename,
                                            shape = shape,
                                            dtype = dtype)
    if type(dtype) is str:
        dt = np.dtype(dtype)
    else:
        dt = dtype
    if nframes is None:
        # Get the number of samples from the file size
        nframes = int(os.path.getsize(filename)/(np.prod(shape)*dt.itemsize))
    dt = np.dtype(dtype)
    return np.memmap(filename,
                     mode=mode,
                     dtype=dt,
                     shape = (int(nframes),*shape))

def load_binary_block(block, #(filename,onset,offset)
                      shape = None,
                      dtype='uint16'): 
    '''Loads a block from a binary file (nchannels,W,H)'''
    fname,offset,bsize = block
    nchans,W,H = shape
    framesize = int(nchans*W*H)
    dt = np.dtype(dtype)
    offset = int(offset)
    with open(fname,'rb') as fd:
        fd.seek(offset*framesize*int(dt.itemsize))
        buf = np.fromfile(fd,dtype = dt, count=framesize*bsize)
    buf = buf.reshape((-1,nchans,W,H),
                               order='C')
    return buf

def compute_trial_baseline_from_binary(trial_onset,
                                       filename,
                                       shape,
                                       nbaseline_frames,
                                       dtype='uint16'):
    dd = load_binary_block((filename,trial_onset,nbaseline_frames),
                           shape=shape,dtype=dtype)
    return dd.mean(axis=0)

def frames_average_for_trials(dat,onsets,nbaseline_frames):
    from .utils import runpar
    if hasattr(dat,'filename'):
        dims = dims = dat.shape[1:]
        dat_path = dat.filename
        frames_average = runpar(compute_trial_baseline_from_binary,onsets,
                                filename = dat_path,
                                shape=dims,
                                nbaseline_frames=nbaseline_frames,
                                dtype = dat.dtype)
    else:
        frame_averages = []
        for on in tqdm(onsets):
            frame_averages.append(dat[on:on+nbaseline_frames].mean(axis=0))
    return np.stack(frames_average)


#######################################################################
################        PARSE WIDEFIELDIMAGER       ###################
#######################################################################

def read_mj2_frames(fname):
    from skvideo.io import FFmpegReader
    sq = FFmpegReader(fname,outputdict={'-pix_fmt':'gray16le'})
    imgs = []
    for s in sq:
        imgs.append(s)
    sq.close()
    return np.stack(imgs).squeeze()

def read_imager_analog(fname):
    from struct import unpack
    with open(fname,'rb') as fd:
        tstamp = unpack("d", fd.read(8))[0]
        onset = unpack("d", fd.read(8))[0]
        nchannels = int(unpack("<d", fd.read(8))[0])
        nsamples = unpack("<d", fd.read(8))[0]
        dat = np.fromfile(fd,dtype='uint16')
        dat = dat.reshape((-1,nchannels)).T
    return dat,dict(baseline=tstamp,
                    onset=onset,
                    nchannels=nchannels,
                    nsamples=nsamples)

def _imager_parse_file(fname, version = 2):
    f,ext = os.path.splitext(fname)
    if ext == '.mj2':
        stack = read_mj2_frames(fname)
    else:
        stack = mmap_dat(fname)
    
    stack = stack.reshape([-1,*stack.shape[-2:]])
    stack = stack[:,100,:]
    stacksize = len(stack)
    _,_,fnum = _parse_binary_fname(fname)
    folder = os.path.dirname(fname)
    analog,analogheader = read_imager_analog(pjoin(folder,'Analog_{0}.dat'.format(fnum[0])))
    idxch1,idxch2,info = _imager_split_channels(stack,analog,analogheader,version = version)
    info['nrecorded'] = len(stack)
    del stack
    return idxch1,idxch2,info

def _imager_split_channels(stack_slice,analog,analogheader,version = 2):
    ''' splits channels from the imager '''
    from .utils import analog_ttl_to_onsets
    #fname_analog = fname_mj2.replace('Frames_','Analog_').replace('.mj2','.dat')
    #stack = read_mj2_frames(fname_mj2)
    #dat,header = read_imager_analog(fname_analog)

    def parse_channels(analog,version = version):
        if version == 1:
            # then the channels are -2 and -1
            ch1,ch1_ = analog_ttl_to_onsets(analog[-2,:],time=None) # this is the blue LED
            ch2,ch2_ = analog_ttl_to_onsets(analog[-1,:],time=None) # this is the violet LED
            ch3_onset,ch3_offset = analog_ttl_to_onsets(analog[3,:],time=None) # this is the stim
            ch4_onset,ch4_offset = analog_ttl_to_onsets(analog[4,:],time=None) # this is another sync
        else:
            ch1,ch1_ = analog_ttl_to_onsets(analog[1,:],time=None) # this is the blue LED
            ch2,ch2_ = analog_ttl_to_onsets(analog[2,:],time=None) # this is the violet LED
            ch3_onset,ch3_offset = analog_ttl_to_onsets(analog[3,:],time=None) # this is the stim
            ch4_onset,ch4_offset = analog_ttl_to_onsets(analog[4,:],time=None) # this is another sync
        return ch1,ch1_,ch2,ch2_,ch3_onset,ch3_offset,ch4_onset,ch4_offset
    ch1,ch1_,ch2,ch2_,ch3_onset,ch3_offset,ch4_onset,ch4_offset = parse_channels(analog,version = version)

    info = dict(baseline = analogheader['baseline'],
                ch1 = ch1,
                ch2 = ch2,
                ch3_onset = ch3_onset,
                ch3_offset = ch3_offset,
                ch4_onset = ch4_onset,
                ch4_offset = ch4_offset,
                onset = analogheader['onset'])
    nframes = stack_slice.shape[0]
    avgnorm = stack_slice.reshape((nframes,-1))
    avgnorm = avgnorm.mean(axis=1)
    avgnorm -= np.mean(avgnorm)/2
    # find the last frame with the LED on; that's ch0
    idx = np.where(avgnorm>0)[0][-1]
    # Find out how many channels were used
    if len(ch1) < stack_slice.shape[0]/2-20 or len(ch2) < stack_slice.shape[0]/2-20:
        print(len(ch1),stack_slice.shape[0]/2-20)
        return np.arange(nframes)[avgnorm>0],[],info
    # otherwise using 2 channels
    # Check if the last frame is odd or even
    if idx & 1:
        chAidx = np.arange(nframes)[1::2]
        chBidx = np.arange(nframes)[0::2]
    else:
        chAidx = np.arange(nframes)[0::2]
        chBidx = np.arange(nframes)[1::2]
    # drop empty frames
    chAidx = chAidx[avgnorm[chAidx]>0]
    chBidx = chBidx[avgnorm[chBidx]>0]
    # look at the analog channels to know which was the last channel.
    lastch = np.argmax([info['ch1'][-1],info['ch2'][-1]])
    if lastch == 1: # last channel was channel 2
        ch1idx = chBidx
        ch2idx = chAidx
    else:
        ch1idx = chAidx
        ch2idx = chBidx
    # collect an equal number of frames for both channels, ch1 first
    if ch1idx[0] > ch2idx[0]:
        # drop first ch2
        ch2idx = ch2idx[1:]
        info['ch2'] = info['ch2'][1:]         
    if ch1idx[-1] > ch2idx[-1]:
        # drop last ch1
        ch1idx = ch1idx[:-1]
        info['ch1'] = info['ch1'][:-1] 
    return ch1idx,ch2idx,info

def parse_imager_mj2_folder(folder, destination, 
                            nchannels = 2,
                            chunksize = 1,
                            dtype = 'uint16'):
    fnames_mj2 = natsorted(glob(pjoin(folder,'Frames_*.mj2')))
    if not len(fnames_mj2):
        print('Folder empty or not accessible {0}'.format(folder))
    sample = read_mj2_frames(fnames_mj2[0])
    nframes,w,h = sample.shape
    
    # cast to float32 so that we can subract the mean (this is sort of dumb..)
    dat_path = pjoin(destination,'frames_{0}_{1}_{2}_{3}.dat'.format(nchannels,w,h,dtype))

    if not os.path.isdir(destination):
        print('Creating output directory {0}'.format(destination))
        os.makedirs(destination)

    tstart = time.time()
    frametrial = [[0,0,0]]
    framesinfo = []
    cnt_trials = 0
    frames_avg = np.zeros((nchannels,w,h),dtype=float).squeeze()
    fnames_mj2_chunks = fnames_mj2
    if chunksize > 1: 
        fnames_mj2_chunks = [fnames_mj2[a:b] for a,b in chunk_indices(len(fnames_mj2),chunksize)]
    with open(dat_path,'wb') as dat:
        for itrial,f in tqdm(enumerate(fnames_mj2_chunks),
                             total=len(fnames_mj2_chunks),
                             desc='Collecting imager trials'):
            if chunksize > 1: 
                res = runpar(split_imager_channels,f)
            else:
                res = [split_imager_channels(f)]
            for ch1,ch2,info in res:
                cnt_trials += 1
                if ch1 is None:
                    print('Skipped trial {0} (no frames).'.format(f))
                    continue
                # compute the stimuli onsets
                if not len(info['stim_onset']) and not len(info['stim_offset']):
                    print('Trial {0} had no stimulus; skiped trial.'.format(cnt_trials))
                    continue
                elif not len(info['stim_onset']):
                    stim_onset_frames = 0 # if there is was an offset but no onset, the stim onset is the first frame (this happens on the visual stim case...) 
                else:
                    try:
                        stim_onset_frames = np.where(info['ch1']<info['stim_onset'])[0][-1]
                    except:
                        print('There was an error in the imager: {0}'.format(info['stim_onset']))
                        stim_onset_frames = 0 # if there is was an offset but no onset, the stim onset is the first frame (this happens on the visual stim case...) 
                d = np.stack([ch1,ch2]).transpose([1,0,2,3])
                frames_avg += np.mean(d,axis = 0).squeeze()
                d = d.reshape([-1,w,h])
                lenframes = frametrial[-1][1] +len(d)/2
                # write meta and frames
                frametrial[-1][2] = frametrial[-1][1] + stim_onset_frames
                frametrial.append([cnt_trials,lenframes,lenframes])
                dat.write(d.astype(dtype))
                framesinfo.append(dict(itrial=cnt_trials,**info))
    frames_avg /= float(len(frametrial))
    np.save(pjoin(destination,'frames_average.npy'),
            frames_avg.astype(np.float32))
    tstop = time.time()
    print('Took {0} min to collect data and compute the averages'.format((tstop-tstart)/60))
    # Save trial onset frames
    trialonsets = np.array(frametrial)
    np.save(pjoin(destination,'trial_onsets.npy'),trialonsets[:-1])
    # Save trial information
    from pandas import DataFrame
    trialinfo = DataFrame(framesinfo)
    trialinfo.to_csv(pjoin(destination,'trial_info.csv'))
    return mmap_dat(dat_path), frames_avg, trialonsets,trialinfo


#######################################################################
################        For handling file sequences ###################
#######################################################################

class GenericStack():
    def __init__(self,filenames,extension):
        self.filenames = filenames
        self.fileextension = extension
        self.dims = None
        self.dtype = None
        self.frames_offset = []
        self.files = []
        self.current_fileidx = None
        self.current_stack = None

    def _get_frame_index(self,frame):
        '''
        Finds out in which file some frames are.
        '''
        fileidx = np.where(self.frames_offset <= frame)[0][-1]
        return fileidx,frame - self.frames_offset[fileidx]
    
    def _load_substack(fileidx):
        pass
    
    def _get_frame(self,frame):
        ''' 
        Returns a single frame from the stack.
        '''
        fileidx,frameidx = self._get_frame_index(frame)
        if not fileidx == self.current_fileidx:
            #print('Loading {0}'.format(fileidx))
            self._load_substack(fileidx)
        return self.current_stack[frameidx]

    def __len__(self):
        return self.shape[0]
    
    def __getitem__(self, *args, squeeze = False):
        index  = args[0]
        idx1 = None
        idx2 = None
        if type(index) is tuple: # then look for 2 channels
            if type(index[1]) is slice:
                idx2 = range(index[1].start, index[1].stop, index[1].step)
            else:
                idx2 = index[1]
            index = index[0]
        if type(index) is slice:
            idx1 = range(*index.indices(self.nframes))#start, index.stop, index.step)
        elif type(index) in [int,np.int32, np.int64]: # just a frame
            idx1 = [index]
        else: # np.array?
            idx1 = index
        img = np.empty((len(idx1),*self.dims),dtype = self.dtype)
        for i,ind in enumerate(idx1):
            img[i] = self._get_frame(ind)
        if not idx2 is None:
            if squeeze:
                return img[:,idx2].squeeze()
            else:
                return img[:,idx2]
        if squeeze:
            return img.squeeze()
        else:
            return img

    def export_binary(self, foldername,
                      basename = 'frames',
                      chunksize = 512,
                      start_frame = 0,
                      end_frame = None,
                      channel = None):
        '''
        Exports a binary file.
        '''
        nframes,nchan,h,w = self.shape
        if end_frame is None :
            end_frame = nframes
        nframes = end_frame - start_frame
        chunks = chunk_indices(nframes,chunksize)    
        chunks = [[c[0]+start_frame,c[1]+start_frame] for c in chunks]
        shape = [nframes,*self.shape[1:]]
        if not channel is None:
            shape[1] = 1
        fname = pjoin('{0}'.format(foldername),'{4}_{0}_{1}_{2}_{3}.bin'.format(
            *shape[1:],self.dtype,basename))
        if not os.path.isdir(os.path.dirname(fname)):
            os.makedirs(os.path.dirname(fname))
        out = np.memmap(fname,
                        dtype = self.dtype,
                        mode = 'w+',
                        shape=tuple(shape))
        for c in tqdm(chunks, desc='Exporting binary'):
            if channel is None:
                out[c[0] - start_frame:c[1] - start_frame] = self[c[0]:c[1]]
            else:
                out[c[0] - start_frame:c[1] - start_frame,0] = self[c[0]:c[1],channel]
        out.flush()
        del out

    def export_tiffs(self, foldername,
                     basename = 'frames',
                     chunksize = 512,
                     start_frame = 0,
                     end_frame = None,
                     channel = None):
        '''
        Exports tifffiles.
        '''
        nframes,nchan,h,w = self.shape
        if end_frame is None :
            end_frame = nframes
        nframes = end_frame - start_frame
        chunks = chunk_indices(nframes,chunksize)    
        chunks = [[c[0]+start_frame,c[1]+start_frame] for c in chunks]
        shape = [nframes,*self.shape[1:]]
        if not channel is None:
            shape[1] = 1

        file_no = 0
        fname = pjoin('{0}'.format(foldername),'{0}_{1:05d}.tif')
        if not os.path.isdir(os.path.dirname(fname)):
            os.makedirs(os.path.dirname(fname))
        from tifffile import imsave
        for c in tqdm(chunks, desc='Exporting tiffs'):
            if channel is None:
                imsave(fname.format(basename,file_no),self[c[0]:c[1]].reshape((-1,*self.dims[1:])))
            else:
                imsave(fname.format(basename,file_no),self[c[0]:c[1],channel].squeeze())
            file_no += 1

        
class ImagerStack(GenericStack):
    def __init__(self,filenames,
                 extension = '.dat',
                 version = 2,# this is because the triggers number changed between the new and the old version...
                 rotate_array=True,
                 imager_preview = False):
        '''
        
        rotate_array=True is for rotating the files saved by the imager...
        '''
        self.rotate_array = rotate_array
        self.fileformat = 'binary'
        self.version = version
        self.extension = extension
        if type(filenames) is str:
            # check if it is a folder
            if os.path.isdir(filenames):
                dirname = filenames
                filenames = natsorted(glob(pjoin(dirname,'Frames*' + self.extension)))
                if not len(filenames): # try mj2's
                    self.extension = '.mj2'
                    filenames = natsorted(glob(pjoin(dirname,'Frames*' + self.extension)))
                    if len(filenames):
                        self.fileformat = 'mj2'
                        self.rotate_array = False # This is not needed for these files...
                    else:
                        raise('Could not find files.')
        super(ImagerStack,self).__init__(filenames,extension)
        
        self.index_ch1 = []
        self.index_ch2 = []
        self.extrainfo = []
        if imager_preview:
            print('Loading only a subset of the frames (preview mode)')
            self.filenames = self.filenames[:3]
        for f in tqdm(self.filenames,desc='Parsing files to access the stack size'):
            # Parse all analog files and frames
            ch1,ch2,info = _imager_parse_file(f,version = self.version)
            self.index_ch1.append(ch1)
            self.index_ch2.append(ch2)
            self.extrainfo.append(info)
        # offset for each file
        self.frames_offset = np.hstack([0,np.cumsum([len(x) for x in self.index_ch1])])
        # get the dims from the first binary file
        if self.fileformat == 'mj2':
            stack = read_mj2_frames(f)
        else:
            stack = mmap_dat(f)
        if self.rotate_array: # fix imager rotation
            if len(stack.shape) == 3:
                stack = stack.transpose([0,2,1])
            if len(stack.shape) == 4:
                stack = stack.transpose([0,1,3,2])
        self.dims = stack.shape[1:]
        # count the number of channels
        self.nchannels = np.sum([len(self.index_ch1[0])>0,
                                 len(self.index_ch2[0])>0])
        if len(self.dims) == 2:
            self.dims = [self.nchannels,*self.dims]

        self.dtype = stack.dtype
        self.nframes = self.frames_offset[-1]
        self.shape = tuple([self.nframes,*self.dims])
        
    def _load_substack(self,fileidx,channel = None):
        if self.fileformat == 'mj2':
            tmp = read_mj2_frames(self.filenames[fileidx])
        else:
            tmp = load_dat(self.filenames[fileidx])
        if self.rotate_array: # fix imager rotation
            if len(tmp.shape) == 3:
                tmp = tmp.transpose([0,2,1])
            if len(tmp.shape) == 4:
                tmp = tmp.transpose([0,1,3,2])
        tmp = tmp.reshape([-1,*tmp.shape[-2:]])
        # combine the indexes from the 2 channels
        if self.nchannels == 2:
            idx = np.sort(np.hstack([self.index_ch1[fileidx],
                                     self.index_ch2[fileidx]]))
        else:
            idx = np.sort(self.index_ch1[fileidx])
            
        try:
            self.current_stack = tmp[idx].reshape([-1,*self.dims])
        except:
            raise OSError('There is a chance that file {0} is corrupt.'.format(
                self.filenames[fileidx]))
        self.current_fileidx = fileidx
                    
class BinaryStack(GenericStack):
    def __init__(self,filenames,
                 extension = '.bin'): # this will try this extension first and then .dat
                
        '''
        Select a stack from a binary file or mutliple binary files
        The file name format needs to ends in _NCHANNELS_H_W_DTYPE.extension

        '''
        self.fileformat = 'binary'
        self.extension = extension
        if type(filenames) is str:
            # check if it is a folder
            if os.path.isdir(filenames):
                dirname = filenames
                filenames = []
                filenames = natsorted(glob(pjoin(dirname,'*'+self.extension)))
                if not len(filenames): # try .dat
                    self.extension = '.dat'
                    filenames = natsorted(glob(pjoin(dirname,'*'+self.extension)))
        if not len(filenames):
            raise(OSError('Could not find files.'))
                
        super(BinaryStack,self).__init__(filenames,extension)
        offsets = [0]
        for f in tqdm(self.filenames, desc='Parsing files to know the stack size'):
            # Parse all binary files
            tmp =  mmap_dat(f)
            dims = tmp.shape[1:]
            dtype = tmp.dtype
            offsets.append(tmp.shape[0])
            del tmp
        # offset for each file
        self.frames_offset = np.cumsum(offsets)
        if nchannels is None:
            nchannels = 1
        self.frames_offset = (self.frames_offset/nchannels).astype(int)
        self.dims = dims[1:]
        if len(self.dims) == 2:
            self.dims = [nchannels,*self.dims]
        self.dims[0] = nchannels

        self.dims = dims
        self.dtype = dtype
        self.nframes = self.frames_offset[-1]
        self.shape = tuple([self.nframes,*self.dims])
        
    def _load_substack(self,fileidx,channel = None):
        self.current_stack = mmap_dat(self.filenames[fileidx])
        self.current_fileidx = fileidx

class TiffStack(GenericStack):
    def __init__(self,filenames,
                 extension = '.tiff', # this will try this extension first and then .tif, .TIFF and .TIF
                 nchannels = 2): 
        '''
        Select a stack from a sequence of TIFF stack files

        '''
        self.extension = extension
        if type(filenames) is str:
            # check if it is a folder
            if os.path.isdir(filenames):
                dirname = filenames
                filenames = []
                for extension in [self.extension,'.tif','.TIFF','.TIF']:
                    if not len(filenames): # try other
                        self.extension = extension
                        filenames = natsorted(glob(pjoin(dirname,'*'+self.extension)))
        if not len(filenames):
            raise(OSError('Could not find files.'))
        super(TiffStack,self).__init__(filenames,extension)
        from tifffile import imread, TiffFile
        self.imread = imread
        self.TiffFile = TiffFile
        offsets = [0]
        for f in tqdm(self.filenames, desc='Parsing tiffs'):
            # Parse all files in the stack
            tmp =  TiffFile(f)
            # get the size from the pages (works with OEM files)
            dims = [len(tmp.pages),*tmp.pages[0].shape]
            #dims = [*tmp.series[0].shape]
            if len(dims) == 2: # then these are single page tiffs
                dims = [1,*dims]
            dtype = tmp.pages[0].dtype
            offsets.append(dims[0])
            del tmp
        # offset for each file
        self.frames_offset = np.cumsum(offsets)
        if nchannels is None:
            nchannels = 1
        self.frames_offset = (self.frames_offset/nchannels).astype(int)
        self.dims = dims[1:]
        if len(self.dims) == 2:
            self.dims = [nchannels,*self.dims]
        self.dims[0] = nchannels
        self.dtype = dtype
        self.nframes = self.frames_offset[-1]
        self.shape = tuple([self.nframes,*self.dims])
        
    def _imread(self, filename):
        arr = None
        with self.TiffFile(filename) as tf:
            arr = np.stack([p.asarray() for p in tf.pages])
        return arr
    
    def _load_substack(self,fileidx,channel = None):
        self.current_stack = self._imread(self.filenames[fileidx]).reshape([-1,*self.dims])
        self.current_fileidx = fileidx

class VideoStack(GenericStack):
    def __init__(self,filenames,
                 extensions = ['.avi','.mov','.mj2'], # this will try this extension first and then .tif, .TIFF and .TIF
                 extension = None,
                 nchannels = None): 
        '''
        Select a stack from a sequence of mov stack files

        '''
        self.extension = extension
        if type(filenames) is str:
            # check if it is a folder
            if os.path.isdir(filenames):
                dirname = filenames
                filenames = []
                for extension in extensions:
                    if not len(filenames): # try other
                        self.extension = extension
                        filenames = natsorted(glob(pjoin(dirname,'*'+self.extension)))
        if not len(filenames):
            raise(OSError('Could not find files.'))
        super(VideoStack,self).__init__(filenames,extension)
        from skvideo.io import FFmpegReader
        self.reader = FFmpegReader
        offsets = [0]
        for fname in self.filenames:
            # Parse all files in the stack
            with FFmpegReader(fname) as f:
                dims = f.getShape()[:-1]
                if f.pix_fmt == 'gray16le':
                    dtype = 'uint16' 
                else:
                    dtype = 'uint8'
                self.pix_fmt = f.pix_fmt 
                offsets.append(dims[0])
                self.framerate = f.inputfps
        # offset for each file
        self.frames_offset = np.cumsum(offsets)
        if nchannels is None:
            nchannels = 1
        self.frames_offset = (self.frames_offset/nchannels).astype(int)
        self.dims = dims[1:]
        self.dims = [nchannels,*self.dims]
        self.dtype = dtype
        self.nframes = self.frames_offset[-1]
        self.shape = tuple([self.nframes,*self.dims])
        self.current_fileidx = -1
        self.current_frameidx = 0
    def _load_substack(self,fileidx,frameidx=0):
        inputdict = {'-pix_fmt':self.pix_fmt}
        outputdict = inputdict
        tidx = (frameidx*self.dims[0])/self.framerate
        t = time.strftime("%H:%M:%S", time.gmtime(tidx))+'{0:.3f}'.format(tidx % 1)[1:]
        self.current_stack = self.reader(self.filenames[fileidx], 
                                          inputdict = {'-ss':t},
                                          outputdict=outputdict)
        self.current_frameidx = frameidx
        self.current_fileidx = fileidx

    def _get_frame(self,frame):
        ''' 
        Returns a single frame from the stack.
        '''
        fileidx,frameidx = self._get_frame_index(frame)
        if not fileidx == self.current_fileidx:
            self._load_substack(fileidx,frameidx)
        elif not frameidx == self.current_frameidx:
            self._load_substack(fileidx,frameidx)        
        for frame in self.current_stack:
            self.current_frameidx = frameidx+1
            return frame.transpose([-1,0,1])

        
def load_stack(foldername, nchannels=None, imager_preview = False):
    ''' 
    Searches the correct format to load from a folder.
    '''
    #    TODO: follow a specific order
    # order = ['binary','tiff','imager','video']
    # First check whats in the folder
    if os.path.isfile(foldername):
        if foldername.endswith('.bin') or foldername.endswith('.dat'): 
            return mmap_dat(foldername)
    # Check binary sequence.
    files = natsorted(glob(pjoin(foldername,'*.bin')))
    if len(files):
        # these don't need channel number because it is written with the filename
        if len(files) == 1:
            return mmap_dat(files[0])
        print('Loading binary stack.')
        return BinaryStack(files) 
    # check tiff sequence
    for ext in ['.TIFF','.TIF','.tif','.tiff']:
        files = natsorted(glob(pjoin(foldername,'*'+ext)))
        if len(files):
            return TiffStack(files, nchannels = nchannels)
    # check imager
    files = natsorted(glob(pjoin(foldername,'Analog*.dat')))
    if len(files):
        return ImagerStack(foldername, imager_preview = imager_preview)
    # check for avi and mov
    for ext in ['.avi','.mov','.mj2']:
        files = natsorted(glob(pjoin(foldername,'*'+ext)))
        if len(files):
            return VideoStack(files, extension = ext, nchannels = nchannels)
    # check for dat
    files = natsorted(glob(pjoin(foldername,'*.dat')))
    if len(files):
        if len(files) == 1:
            return mmap_dat(files[0])
        return BinaryStack(foldername)

